/** $Header: /p/condor/repository/CONDOR_SRC/src_java/condor/classad/AttrRef.java,v 1.11 2002/07/20 21:21:28 roy Exp $ */

/*--------------------------------------------------------------------
 *
 * Condor Java ClassAd library
 * Copyright (C) 1990-2001, CONDOR Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and Marvin Solomon
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General
 * Public License as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *--------------------------------------------------------------------*/

package condor.classad;

import java.util.*;

/**
 * A reference to another attribute.
 * A leaf node in an expression tree that represents an attribute reference.
 * It includes a ``locality'' (one of SELF, PARENT, SUPER, or ROOT) and
 * an optional pointer to a target RecordExpr.  The target is set at compile
 * time by a post-pass over the expression tree as follows
 * <dl>
 * <dt>SELF<dd>The target is smallest enclosing RecordExpr.  The concrete
 * syntax "xxx" is compiled as SELF.xxx (i.e., Select(AttrReff(SELF),"xxx")).
 * <dt>ROOT<dd>The target is the largest enclosing RecordExpr.  The concrete
 * syntax ".xxx" is compiled as ROOT.xxx.
 * <dt>PARENT<dd>The target is the second smallest enclosing RecordExpr.  In
 * other words, the smallest RecordExpr properly enclosing the RecordExpr
 * denoted by SELF.
 * <dt>SUPER<dd>If the SELF RecordExpr does not have any attribute named
 * "super", the target is the same as PARENT; else if the super attribute of the
 * SELF RecordExpr is an AttrRef, the target is the same as the target of that
 * AttrRef; otherwise, it is null.
 * </dl>
 * The target is null if the indicated RecordExpr does not exist.  In
 * particular, for SELF and ROOT, the target is null if this AttrRef is not
 * nested inside any RecordExpr.  For PARENT, it is null if it is not nested at
 * least two deep inside RecordExprs.
 * If the target is null, it is dynamically supplied at
 * evaluation time from an associated Context.
 * @see Expr
 * @see RecordExpr
 * @see Context
 * @author <a href="mailto:solomon@cs.wisc.edu">Marvin Solomon</a>
 * @version 1.2
 */
public class AttrRef extends Expr {
    private static String VERSION = "$Id: AttrRef.java,v 1.11 2002/07/20 21:21:28 roy Exp $";

    /** The expression referenced by this AttrRef, or null.  */
    private RecordExpr target;

    /** The scope of this reference:  one of SELF, ROOT, PARENT, or SUPER.
     * @see #SELF
     * @see #ROOT
     * @see #PARENT
     * @see #SUPER
     */
    public final int locality;

    /** Table for translating locality values to strings.  */
    private final static String[] localityName = {
        "this", "root", "parent", "super" };

    /** A value for locality indicating that this is a reference to the
     * immediately enclosing classad.
     */
    public static final int SELF = 0;

    /** A value for locality indicating that this is a reference to the
     * outermost enclosing classad.
     */
    public static final int ROOT = 1;

    /** A value for locality indicating that this is a reference to the
     * second smallest enclosing classad.
     */
    public static final int PARENT = 2;

    /** A value for locality indicating that this is a reference to a "super"
     * classad.
     */
    public static final int SUPER = 3;

    /** Create a new AttrRef.
     * @param locality one of SELF, ROOT, PARENT, or SUPER.
     */
    public AttrRef(int locality) {
        super(ATTRIBUTE);
        this.locality = locality;
        this.target = null;
    } // AttrRef(int)

    /** Convert this Expr to a string.
     * The exact format of the result is controled by the static fields
     * <code>Expr.formatFlags</code> and <code>Expr.formatIndent</code>.
     * The result may be multiple lines long (it may include embedded newlines
     * and tabs), but it never ends with a newline.
     * @return a string representation of the this AttrRef
     * @see Expr#formatFlags
     * @see Expr#formatIndent
     */
    public String toString() {
        String result = localityName[locality];
        if (target != null)
            result += ":" + target.instance;
        return result;
    } // toString()

    /** Make a recursive traversal over an Expr, filling in targets of AttrRefs
     * and parent links of RecordExprs.
     * @param e the root of the Expr to be traversed.
     */
    public static void setTargets(Expr e) {
        if (e == null)
            return;
        if (e.type == RECORD)
            setTargets(e, (RecordExpr)e, null, null);
        else
            setTargets(e, null, null, null);
    } // setTargets(Expr)

    /** Make a recursive traversal over an Expr, filling in targets of AttrRefs
     * and parent links of RecordExprs.
     * @param e the root of the Expr to be traversed.
     * @param parent the smallest enclosing RecordExpr, if any.
     * @param grandParent the second smallest enclosing RecordExpr, if any.
     */
    private static void setTargets(Expr e,
        RecordExpr root, RecordExpr parent, RecordExpr grandParent)
    {
        if (e == null)
            return;
        switch (e.type) {
            default: break;
            case COND:
                setTargets(((CondExpr)e).ec, root, parent, grandParent);
                setTargets(((CondExpr)e).et, root, parent, grandParent);
                setTargets(((CondExpr)e).ef, root, parent, grandParent);
                break;
            case OP:
                setTargets(((Op)e).arg1, root, parent, grandParent);
                setTargets(((Op)e).arg2, root, parent, grandParent);
                break;
            case CALL: {
                Expr[] args = ((FuncCall)e).args;
                for (int i=0; i < args.length; i++)
                    setTargets(args[i], root, parent, grandParent);
                break;
            }
            case SELECTION:
                setTargets(((SelectExpr)e).base, root, parent, grandParent);
                break;
            case SUBSCRIPT:
                setTargets(((SubscriptExpr)e).base, root, parent, grandParent);
                setTargets(((SubscriptExpr)e).selector,
                    root, parent, grandParent);
                break;
            case ATTRIBUTE: {
                AttrRef a = (AttrRef)e;
                switch (a.locality) {
                    case SELF: a.target = parent; break;
                    case ROOT: a.target = root; break;
                    case PARENT: a.target = grandParent; break;
                    default: break; // can't handle SUPER (yet)
                }
                break;
            }
            case LIST:
                for (Iterator i = ((ListExpr)e).iterator(); i.hasNext(); ) {
                    setTargets((Expr) i.next(), root, parent, grandParent);
                }
                break;
            case RECORD: {
                grandParent = parent;
                parent = (RecordExpr)e;
                for (Iterator val = parent.values(); val.hasNext();) {
                    setTargets((Expr)val.next(), root, parent, grandParent);
                }
            }
        }
    } // setTargets(Expr,RecordExpr,RecordExpr,RecordExpr)

    /** Evalutate this expression.
     * @param ctx an evaluation context for evaluating this expression
     * @return a Constant, ListExpr, or RecordExpr representing the value.
     * @see Expr#eval(condor.classad.Context)
     */
    protected Expr reEval(Context ctx) {
        if (target != null) {
            return target;
        }
        switch (locality) {
            case SELF: return Constant.Undef;
            case ROOT: return ctx.root;
            case PARENT: return ctx.root;
            case SUPER:
        }
        if (target == null) {
            if (locality != ROOT)
                throw new RuntimeException(
                    "attempt to evaluate a relative attribute reference with"
                    + " no static scope");
            return ctx.root;
        }
        return target;
    } // reEval(Context)

    /** The precedence of this expression
     * @return the precendence of this node (MAXPREC).
     */
    public int prec() {
        return MAXPREC;
    } // prec()
}
