/** $Header: /p/condor/repository/CONDOR_SRC/src_java/condor/classad/CiString.java,v 1.4 2002/07/20 21:21:28 roy Exp $ */

/*--------------------------------------------------------------------
 *
 * Condor Java ClassAd library
 * Copyright (C) 1990-2001, CONDOR Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and Marvin Solomon
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General
 * Public License as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *--------------------------------------------------------------------*/


package condor.classad;
import java.util.*;
import java.io.*;

/** A "case-ignore" string that behaves just like a String except for
 * its equals, compareTo, and hasCode methods, which ignore case.  This is not
 * a general-purpose implementation but one specifically optimized for use
 * as a key in a Map.  For example, it assumes that nobody will ever
 * try to compare a CiString to anything other than another CiString (and
 * will throw a ClassCastException or NullPointerException if this
 * assumption is violated).
 * <p> The implementation stores both the actual String value and its
 * lower-case translation.  It uses a static mapping from mixed-case to
 * lower-case values to ensure that there is at most one instance of each
 * kind of string in the program.  This technique provides advantages in
 * both speed (comparison is just pointer equality) and space.  A typical
 * application will have many classads with the same set of attributes, and
 * typically most or all of the instances of an attribute will be spelled
 * with the same capitalization.
 * <p> This class implements Comparable so that RecordExpr's can be
 * represented as TreeMaps, which have much lower space overhead than
 * HashMaps.  However, the ordering is based on where the address in memory of
 * the intern'ed lower-case strings rather lexicographic ordering.
 */
class CiString implements Comparable {
    private static String VERSION = "$Id: CiString.java,v 1.4 2002/07/20 21:21:28 roy Exp $";

    /** The actual String being encapsulated. */
    private String pname;

    /** The lower-case version of pname, as returned by
     * pname.toLowerCase().intern().  Thus all "equal" instances (instances
     * that have equal lower-case versions) share the same value of lc.
     */
    private String lc;

    /** Constructor.  Note that it is private.  Clients should use
     * the getInstance() method rather than new.
     * @param v the encapsulated String.
     */
    private CiString(String v) {
        pname = v.intern();
        lc = v.toLowerCase().intern();
    } // CiString(String)

    /** Convert to String.
     * @return the string in its original capitalization.
     */
    public String toString() {
        return pname;
    } // toString()

    /** Compare this CiString to another CiString.
     * Since there is a unique instance for any lower-case value,
     * the comparison is simply pointer equality.
     * @param o the other CiString
     * @return true if the strings are equal, ignoring case.
     * @exception ClassCastException if o is not an instance of CiString.
     * @exception NullPointerException if o is null.
     */
    public boolean equals(Object o) {
        return lc == ((CiString) o).lc;
    } // equals

    /** Compare this CiString to another CiString.
     * The result is a consistent ordering compatible with equals, but is not
     * based on anything reasonable such as lexicographic ordering.  It is only
     * intended to be used to allow CiStrings to be used as keys in SortedMaps.
     * @param o the other CiString
     * @return  negative integer, zero, or a positive integer as this
     *          CiString preceeds, matches, or follows the other CiString
     *          in an arbitrary ordering.
     * @exception ClassCastException if o is not an instance of CiString.
     * @exception NullPointerException if o is null.
     */
    public int compareTo(Object o) {
        return 
            System.identityHashCode(lc)
                - System.identityHashCode(((CiString)o).lc);
    } // compareTo

    /** Compute a hash code for this string.  Strings that are equal except
     * for case return the same hash code.  Since there is a unique
     * instance for any lower-case value, the result is simply the
     * "address" of this CiString, as returned by
     * System.identityHashCode(this).
     * @return the hash code.
     */
    public int hashCode() {
        return System.identityHashCode(this);
    } // hashCode

    /** A mapping from String values to instances of CiString.
     * If instanceMap.get(s) is an instance c of CiString such
     * that s.toLowerCase().equals(c.lc).  It is the unique instance such
     * that c.pname == s.
     * Note that original-case strings are stored only once and shared as
     * keys in instanceMap and values of c.pname.
     */
    private static Map instanceMap = new HashMap();

    /** Returns the unique instance c of CiString such that
     * c.pname.equals(v).  Creates such an instance if necessary.
     * @param v the mixed-case value.
     * @return the corresponding CiString instance.
     */
    public static CiString getInstance(String v) {
        if (v == null) {
            return null;
        }
        CiString result = (CiString) instanceMap.get(v);
        if (result == null) {
            result = new CiString(v);
            instanceMap.put(v, result);
        }
        return result;
    } // getInstance(String)
/*
    public static void debug() {
        int len = 0;
        System.out.println(instanceMap.size() + " CiStrings:");
        for (Iterator i = instanceMap.keySet().iterator(); i.hasNext(); ){
            String s = (String)i.next();
            len += s.length();
            System.out.print(" " + s);
        }
        System.out.println();
        System.out.println("total bytes " + len);
    }
*/
} // CiString
