package condor.classad;

import java.util.*;
import java.io.*;

/* $Header: /p/condor/repository/CONDOR_SRC/src_java/condor/classad/ClassAd.java,v 1.7 2002/07/20 21:21:28 roy Exp $ */

/*--------------------------------------------------------------------
 *
 * Condor Java ClassAd library
 * Copyright (C) 1990-2001, CONDOR Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and Marvin Solomon
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General
 * Public License as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *--------------------------------------------------------------------*/

/**
 * A library of handy methods for manipulating classad expressions.
 * The class contains only static methods for performing various operations on
 * classified advertisements.  See {@link Expr} and its various subclasses,
 * especially {@link RecordExpr} for the actual objects that are manipulated.
 * @see Expr
 * @see RecordExpr
 * @author <a href="mailto:solomon@cs.wisc.edu">Marvin Solomon</a>
 * @version 1.2
 */
public class ClassAd {
    private static String VERSION = "$Id: ClassAd.java,v 1.7 2002/07/20 21:21:28 roy Exp $";

    // Convenience functions for matching class ads.

    /** The default constructor is made private to prevent anybody from trying
     * to create an instance of this class.
     */
    private ClassAd() {
        throw new RuntimeException();
    } // ClassAd()

    /** Wrap a pair of ads in an environment and context such that inside
     * ad1, <code>self</code> refers to <code>ad1</code>, <code>other</code>
     * refers to <code>ad2</code>, and any other attribute is searched for
     * first in <code>ad1</code> and then in <code>ad2</code> (and similarly,
     * with <code>ad1</code> and <code>ad2</code> interchanged).
     * @param ad1 the first ClassAd.
     * @param ad2 the second ClassAd.
     * @return a <code>Context ctx</code> such that
     * <code>E.Ad1.self</code> is <code>ad1</code>, and
     * <code>E.Ad2.self</code> is <code>ad2</code>, where <code>E =
     * ctx.root</code>
     */
     public static Context bind(RecordExpr ad1, RecordExpr ad2) {
        // Create the classad env =
        // [
        //    Ad1 = [other = .Ad2.self; self = <ad1>]
        //    Ad2 = [other = .Ad1.self; self = <ad2>]
        // ]
        // Note that env1 and env2 (the values of the Ad1 and Ad2 attributes)
        // are created with null parent pointers.
        // It is also assumed that <ad1> and <ad2> have null parent pointers.
        // The Context created below wires these four ads together with dynamic
        // "parent pointers".
        // In RecordExpr.search(String name, Context ctx),
        // the search uses ctx to find parent pointers if the "compiled in"
        // parent pointer is null.
        RecordExpr env = new RecordExpr();
        RecordExpr env1 = new RecordExpr();
        env1.insertAttribute(
                "other",
                new AttrRef(AttrRef.ROOT)
                    .select("Ad2").select("self"))
            .insertAttribute(
                "self",
                ad1);
        RecordExpr env2 = new RecordExpr();
        env2.insertAttribute(
                "other",
                new AttrRef(AttrRef.ROOT)
                    .select("Ad1").select("self"))
            .insertAttribute(
                "self",
                ad2);
        env.insertAttribute("Ad1", env1);
        env.insertAttribute("Ad2", env2);
        AttrRef.setTargets(env);

        // Create a context containing the parent mappings
        // ad1->env1->ad2->env2->ad1
        Context ctx =
            new Context(env)
                .setParent(ad1,env1)
                .setParent(env1,ad2)
                .setParent(ad2,env2)
                .setParent(env2,ad1);
        return ctx;
    } // bind(RecordExpr,RecordExpr)

    /** Evaluate a selection from a ClassAd.
     * @param ad a ClassAd from which to select.
     * @attrs a sequence <code>attr1, attr2, ..., attrn</code> of attribute
     * names.
     * @return the result of evaluating <code>ad.attr1.attr2...attrn</code>
     * in a context in which absolute attribute reference (ones starting with
     * `.') are evaluated relative to <code>ad</code>.
     */
    public static Expr eval(RecordExpr ad, String[] attrs) {
        Expr sel = ad;
        for (int i=0; i<attrs.length; i++)
            sel = sel.select(attrs[i]);
        Context ctx = new Context(ad);
        return sel.eval(ctx);
    } // eval(RecordExpr,String[])

    /** Evaluate a selection from a ClassAd in a given evaluation context.
     * @param ctx a Context for the evaluation.
     * @attrs a sequence <code>attr1, attr2, ..., attrn</code> of attribute
     * names.
     * @return the result of evaluating <code>ad.attr1.attr2...attrn</code>
     * in a context <code>ctx</code>, where <code>ad = ctx.root</code>.
     */
    public static Expr eval(Context ctx, String[] attrs) {
        Expr sel = ctx.root;
        for (int i=0; i<attrs.length; i++)
            sel = sel.select(attrs[i]);
        return sel.eval(ctx);
    } // eval(Context,String[])

    /** Evaluate an attribute of a ClassAd in an environment consisting of
     * a pair of ads. The call <code>ClassAd.eval(attr, ad1, ad2)</code>
     * evaluates <code>ad1.attr</code> an environment such that inside
     * ad1, <code>self</code> refers to <code>ad1</code>, <code>other</code>
     * refers to <code>ad2</code>, and any other attribute is searched for
     * first in <code>ad1</code> and then in <code>ad2</code> (and similarly,
     * with <code>ad1</code> and <code>ad2</code> interchanged).
     * @param attr the name of the attribute used to initiate the evaluation
     * @param ad1 the ClassAd used to initiate the evaluation
     * @param ad2 the other ClassAd
     * @return the result of the evaluation
     */
    public static Expr eval(String attr, RecordExpr ad1, RecordExpr ad2) {
        return eval(bind(ad1, ad2), new String[] { "Ad1", "self", attr });
    } // eval(String,RecordExpr,RecordExpr)

    /** Match two ClassAds.  If each expr is a RecordExpr with a
     * <code>Requirements</code> attribute that evaluates to <b>true</b> in the
     * context of the other expr and each expr has a <code>Rank</code>
     * attribute that evaluates to an integer, the result is two-element array
     * containing the <code>Rank</code> values.  In all other cases, the result
     * is <b>null</b>.
     * @param expr1 the first expression.
     * @param expr2 the second expression.
     * @return the result of the match, either a pair of integers or
     * <b>null</b>.
     */
    public static int[] twoWayMatch(Expr expr1, Expr expr2) {
        if (expr1.type != Expr.RECORD || expr2.type != Expr.RECORD)
            return null;
        RecordExpr ad1 = (RecordExpr)expr1;
        RecordExpr ad2 = (RecordExpr)expr2;
        Context ctx = bind(ad1, ad2);

        Expr res;
        Constant val;
        int rank1, rank2;

        // Check ad1.requirements
        res = eval(ctx, new String[] { "Ad1", "self", "requirements" });
        if (!res.isConstant())
            return null;
        val = (Constant)res;
        if (!val.toBoolean())
            return null;

        // Check ad2.requirements
        res = eval(ctx, new String[] { "Ad2", "self", "requirements" });
        if (!res.isConstant())
            return null;
        val = (Constant)res;
        if (!val.toBoolean())
            return null;

        // Evaluate ad1.rank
        res = eval(ctx, new String[] { "Ad1", "self", "requirements" });
        if (!res.isConstant())
            return null;
        val = (Constant)res;
        if (val.type != Expr.BOOLEAN)
            return null;
        rank1 = ((Integer)(val.value)).intValue();

        // Evaluate ad2.rank
        res = eval(ctx, new String[] { "Ad2", "self", "requirements" });
        if (!res.isConstant())
            return null;
        val = (Constant)res;
        if (val.type != Expr.BOOLEAN)
            return null;
        rank2 = ((Integer)(val.value)).intValue();

        return new int[]{ rank1, rank2 };
    } // twoWayMatch(Expr,Expr)

    /** Match two ClassAds.  If <code>expr1</code> is a RecordExpr with a
     * <code>Requirements</code> attribute that evaluates to <b>true</b> in the
     * context of <code>expr2</code>, then return <b>true</b>.  Otherwise
     * return <b>false</b>.
     * @param expr1 the first expression.
     * @param expr2 the second expression.
     * @return the result of the match.
     */
    public static boolean oneWayMatch(Expr expr1, Expr expr2) {
        if (expr1.type != Expr.RECORD || expr2.type != Expr.RECORD)
            return false;
        RecordExpr ad1 = (RecordExpr)expr1;
        RecordExpr ad2 = (RecordExpr)expr2;
        Context ctx = bind(ad1, ad2);

        Expr res;
        Constant val;

        // Check ad1.requirements
        res = eval(ctx, new String[] { "Ad1", "self", "requirements" });
        if (!res.isConstant())
            return false;
        val = (Constant)res;
        return val.toBoolean();
    } // oneWayMatch(Expr,Expr)

    // Factories (static methods returning new objects)
    // See also ClassAdParser

    /** Create a constant expression from an integer value.
     * @param i the value.
     * @return the resulting Constant object.
     */
    public static Constant constant(int i) {
        return Constant.getInstance(i);
    } // constant(int)

    /** Create a constant expression from a real value.
     * @param x the value.
     * @return the resulting Constant object.
     */
    public static Constant constant(double x) {
        return Constant.getInstance(x);
    } // constant(double)

    /** Create a constant expression from a String value.
     * @param s the value.
     * @return the resulting Constant object.
     */
    public static Constant constant(String s) {
        return Constant.getInstance(s);
    } // constant(String)
} // ClassAd
