package condor.classad;

import java.util.*;
import java.io.*;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.ParserConfigurationException;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.InputSource;

/* $Header: /p/condor/repository/CONDOR_SRC/src_java/condor/classad/ClassAdParser.java,v 1.10 2002/07/20 21:21:28 roy Exp $ */

/*--------------------------------------------------------------------
 *
 * Condor Java ClassAd library
 * Copyright (C) 1990-2001, CONDOR Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and Marvin Solomon
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General
 * Public License as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *--------------------------------------------------------------------*/

/**
 * A factory for creating classad expressions (instances of {@link
 * condor.classad.Expr}), by parsing a textual or XML representation.
 * Input can come from a String, a Reader, or an InputStream.
 * @see ClassAd
 */
public class ClassAdParser {
    private static String VERSION = "$Id: ClassAdParser.java,v 1.10 2002/07/20 21:21:28 roy Exp $";

    // ========== public fields ===========

    /** Flag indicating that input format is "text".
     * @see #XML
     */
    public static final int TEXT = 0;

    /** Flag indicating that input format is xml.
     * @see #TEXT
     */
    public static final int XML = 1;

    /** Input format.
     * Either {@link #TEXT} or {@link #XML}.
     */
    public final int format;

    // ========== private fields ===========

    /** The underlying parser for text representation (only used if format is
     * TEXT).
     */
    private Parser parser;

    /** The underlying parser for xml representation (only used if format is
     * XML).
     */
    private SAXParser saxParser;

    /** Source of characaters for parsing (on used if format is XML). */
    private InputSource source;

    /** SAX event handler for parsing (on used if format is XML). */
    private ClassAdSAXHandler saxHandler;

    /** Verbosity of error messages. */
    private int verbosity = 3;

    // ========== constructors ===========

    /** Create a parser to parse text input.
     * An input source needs to be supplied by one of the reset methods before
     * it can be used for parsing.
     */
    public ClassAdParser() {
        this(TEXT);
    } // ClassAdParser()

    /** Create a parser to parse input.
     * An input source needs to be supplied by one of the reset methods before
     * it can be used for parsing.
     * @param format either {@link #TEXT} or {@link #XML}.
     */
    public ClassAdParser(int format) {
        this.format = format;
        switch (format) {
        case TEXT:
            break;
        case XML:
            SAXParserFactory factory = SAXParserFactory.newInstance();
            factory.setValidating(true); // TODO: allow caller to decide
            try {
                saxParser = factory.newSAXParser();
            } catch (ParserConfigurationException e) {
                throw new IllegalArgumentException(
                            "Cannot create parser: " + e);
            } catch (SAXException e) {
                throw new IllegalArgumentException(
                            "Cannot create parser: " + e);
            }
            saxHandler = new ClassAdSAXHandler();
            break;
        default:
            throw new IllegalArgumentException("Invalid format " + format);
        }
    } // ClassAdParser()

    /** Create a parser to parse a string in default (text) format.
     * @param s the string to be parsed.
     */
    public ClassAdParser(String s) {
        this(TEXT);
        reset(s);
    } // ClassAdParser(String)

    /** Create a parser to parse a string.
     * @param s the string to be parsed.
     * @param format either {@link #TEXT} or {@link #XML}.
     */
    public ClassAdParser(String s, int format) {
        this(format);
        reset(s);
    } // ClassAdParser(String,int)

    /** Create a parser to parse an InputStream in default (text) format.
     * @param str the string to be parsed.
     */
    public ClassAdParser(InputStream str) {
        this(TEXT);
        reset(str);
    } // ClassAdParser(InputStream)

    /** Create a parser to parse an InputStream.
     * @param str the string to be parsed.
     * @param format either {@link #TEXT} or {@link #XML}.
     */
    public ClassAdParser(InputStream str, int format) {
        this(format);
        reset(str);
    } // ClassAdParser(String,int)

    /** Create a parser to parse characters in default (text) format.
     * @param rdr the string to be parsed.
     */
    public ClassAdParser(Reader rdr) {
        this(TEXT);
        reset(rdr);
    } // ClassAdParser(Reader)

    /** Create a parser to parse characters.
     * @param rdr the string to be parsed.
     * @param format either {@link #TEXT} or {@link #XML}.
     */
    public ClassAdParser(Reader rdr, int format) {
        this(format);
        reset(rdr);
    } // ClassAdParser(Reader,int)

    // ========== public methods ===========

    /** Reset this parser to parse input from a String.
     * @param s the string to parse.
     */
    public void reset(String s) {
        reset(new StringReader(s));
    } // reset(String)

    /** Reset this parser to parse input from an InputStream
     * @param str the stream that supplies the input.
     */
    public void reset(InputStream str) {
        switch (format) {
        case TEXT:
            parser = new Parser(new InputStreamReader(str));
            parser.verbosity = verbosity;
            break;
        case XML:
            source = new InputSource(str);
            break;
        }
    } // reset(InputStream)

    /** Reset this parser to parse input from a Reader.
     * @param rdr the reader that supplies the input.
     */
    public void reset(Reader rdr) {
        switch (format) {
        case TEXT:
            parser = new Parser(rdr);
            parser.verbosity = verbosity;
            break;
        case XML:
            source = new InputSource(rdr);
            break;
        }
    } // reset(Reader)

    /** Parse a classsad Expr from the current input stream.
     * If format is TEXT, the longest possible substring of the input
     * representing a valid expression is parsed.  The stream is left
     * positioned so that the next token to be read is the first token (if any)
     * not part of the expression parsed.  For example, if the input stream
     * contains
     * <pre>
     *    a + b c
     * </pre>
     * The first call to parse() will return "a+b" and a subsequent call
     * will return "c".
     * <p>
     * If format is XML, the entire stream will be parsed as a single Expr.
     * In this case, subsequent calls will always return null.
     *
     * @return the parsed Expr or null on EOF or error.
     * @throws IllegalStateException if no input source has been specified by
     *         a constructor or a call to a reset method.
     * @see #nextToken()
     */
    public Expr parse() {
        Expr result = null;
        switch (format) {
        case TEXT:
            if (parser == null) {
                throw new IllegalStateException(
                            "No input source has been specified");
            }
            result = (Expr)parser.parse();
            break;
        case XML:
            if (source == null) {
                throw new IllegalStateException(
                            "No input source has been specified");
            }
            try {
                saxParser.parse(source, saxHandler);
                result = saxHandler.getResult();
                // A SAX parser only allows one bite at the apple
                source.getCharacterStream().close();
                source = null;
            } catch (Exception e) {
                if (verbosity > 1) {
                    e.printStackTrace();
                } else if (verbosity > 0) {
                    System.err.println(e.getMessage());
                }
            }
            break;
        }
        if (result != null) {
            AttrRef.setTargets(result);
        }
        return result;
    } // parse()

    /** Parse a Query from the current input stream.
     * @return the query
     */
    public Query parseQuery() {
        if (format != TEXT) {
            throw new IllegalStateException(
                        "Query parsing not supported for XML parsers");
        }
        if (parser == null) {
            throw new IllegalStateException(
                        "No input source has been specified");
        }
        return (Query)parser.parse();
    } // parseQuery()

    /** Get the "lookahead" token: the first token not yet consumed by the
     * parser.  The result is an internal code.  In the case of one-charcter
     * tokens, the code is the same as the integer value of the character.
     * The code 0 means end-of-file.
     * The token is not removed from the input stream.
     * For {@link #XML} parsers, always returns 0;
     * @return the code for the next token.
     * @see #getNextToken()
     */
    public int nextToken() {
        return format == TEXT ? parser.nextToken() : 0;
    } // nextToken()

    /** The the "value" of the lookahead token.
     * For {@link #XML} parsers, always returns null;
     * @see #nextToken()
     * @return the value
     */
    public Object nextValue() {
        return format == TEXT ? parser.nextValue() : null;
    } // nextValue()

    /** Get and remove the "lookahead" token: the first token not yet consumed
     * by the parser.  The result is an internal code.  In the case of
     * one-charcter tokens, the code is the same as the integer value of the
     * character.  The code 0 means end-of-file.
     * The token is removed from the input stream.
     * For {@link #XML} parsers, always returns 0;
     * @return the code for the next token.
     * @see #getNextToken()
     */
    public int getNextToken() {
        return format == TEXT ? parser.getNextToken() : 0;
    } // getNextToken()

    /** Control whether the actions of the parser are traced.  The default is
     * not to trace.
     * For {@link #XML} parsers, has no effect.
     * @param on if true, turn on tracing; if false, turn it off.
     * @return the previous value of the tracing flag.
     */
    public boolean enableTracing(boolean on) {
        boolean result = parser.debug;
        parser.debug = on;
        return result;
    } // enableTracing(boolean)

    /** Set verbosity of error messages (always sent to System.err).
     * <dl>
     * <dt>0<dd>
     * <dt>0<dd>no messages
     * <dt>1<dd>only show message and line number
     * <dt>2<dd>also echo line in question
     * <dt>3<dd>also list tokens expected
     * </dl>
     * The default is maximally verbose.  Levels 2 and 3 are only meaningful
     * for {@link #TEXT} format parsers.
     * @param level desired level.
     * @return old level.
     */
    public int setVerbosity(int level) {
        int result = verbosity;
        verbosity = level;
        switch (format) {
        case TEXT:
            parser.verbosity = level;
            break;
        case XML:
            saxHandler.verbosity = level;
            break;
        }
        return result;
    } // setVerbosity
} // ClassAdParser
