package condor.classad;

import java.io.*;
import java.util.Stack;
import java.util.Date;
import java.text.SimpleDateFormat;
import java.text.ParsePosition;

import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.ParserConfigurationException;
import org.xml.sax.SAXParseException;
import org.xml.sax.SAXException;
import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.helpers.DefaultHandler;

/* $Header: /p/condor/repository/CONDOR_SRC/src_java/condor/classad/ClassAdSAXHandler.java,v 1.2 2002/07/20 21:21:28 roy Exp $ */

/*--------------------------------------------------------------------
 *
 * Condor Java ClassAd library
 * Copyright (C) 1990-2001, CONDOR Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and Marvin Solomon
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General
 * Public License as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *--------------------------------------------------------------------*/

/** A handler for SAX (Simple API for XML) events from a SAXParser.
 * It uses those events to build up an Expr object.
 * Note that this class is not public; it may only be referenced within the
 * condor.classad package.  It should only be referenced by XMLParser.
 */
class ClassAdSAXHandler extends DefaultHandler {

    // ******** Internal state ***********

    /** Stack of partially constructed expressions, innermost at top. */
    private Stack stack = new Stack();

    /** Concatenated PCDATA content of current n, s, e, or t element). */
    private StringBuffer pcdata = new StringBuffer();

    /** Result of parsing pcdata */
    private Expr element;

    /** Format for parsing ISO 8601 date and time */
    private SimpleDateFormat dateAndTimeFormat
        = new SimpleDateFormat("yyyyMMdd'T'hhmmss");

    /** Format for parsing ISO 8601 date */
    private SimpleDateFormat dateFormat = new SimpleDateFormat("yyyyMMdd");

    /** Cursor for parsing dates and times. */
    private ParsePosition parsePos = new ParsePosition(0);

    /** The "n" attribute of the current &lt;ca&gt; tag (if any).  */
    private String attrName;

    /** A parser for parsing &lt;e&gt; elements. */
    private ClassAdParser caParser;

    /** Handle on the original document for reporting the locations of errors.
     */
    Locator locator;

    /** Verbosity of error messages. */
    public int verbosity = 3;

    // ******** Public interface ***********

    /** Get the result of the most recent parse, if any.
     * @return the resulting Expr, or null.
     */
    public Expr getResult() {
        if (stack.empty()) {
            return null;
        } else {
            return (Expr) stack.peek();
        }
    } // getResult

    // ******** Overridden DefaultHandler methods ***********

    /** Called by SAXParser to deliver a locator.
     * @param locator An object that can return the location of any SAX
     *                document event.
     */
    public void setDocumentLocator(Locator locator) {
        this.locator = locator;
    } // setDocumentLocator

    /** Called by SAXParser to announce warnings.
     * @param e The warning information encoded as an exception.
     */
    public void warning(SAXParseException e) {
        printMessage("Warning", e);
    } // warning

    /** Called by SAXParser if it discovers a (non-fatal) error. */
    public void error(SAXParseException e) {
        printMessage("Error", e);
    } // error

    /** Called by SAXParser after recognizing some "PCDATA" (parseed character
     * data).  Simply appends the characters to pcdata.
     * @param ch The characters.
     * @param start The start position in the character array.
     * @param length The number of characters to use from the character array. 
     */
    public void characters(char[] ch, int start, int length) {
        pcdata.append(ch, start, length);
    } // charaacters

private static void pl(Object o ) { System.err.println(o); }
    /** Called by SAXParser at the start of an element after parsing the start
     * tag.
     * @param uri currenly not documented; appears always to be ''.
     * @param localName The local name (without prefix), or the empty string if
     *                  Namespace processing is not being performed.  Currently
     *                  ignored.
     * @param name The qualified name (with prefix), or the empty string if
     *                  qualified names are not available.
     * @param attrs The specified or defaulted attributes.  For classad.dtd,
     *              there is at most one attribute per tag.
     */
    public void startElement(String uri,
                             String localName,
                             String name,
                             Attributes attrs)
            throws SAXException
    {
        //Expr.db("start " + name);
        if ("classads".equals(name) || "l".equals(name)) {
            // <classads> is treated just like <l>: Just push and empty
            // ListExpr on the stack.
            stack.push(new ListExpr());
        } else if ("ca".equals(name)) {
            // <ca>:  push an empty RecordExpr on the stack.
            stack.push(new RecordExpr());
        } else if ("a".equals(name)) {
            // <a n="x">:  save "x"
            attrName = attrs.getValue(0);
        } else if ("b".equals(name)) {
            addElement(attrs.getValue(0).charAt(0) == 't'
                        ? Constant.True : Constant.False);
        } else if ("un".equals(name)) {
            addElement(Constant.Undef);
        } else if ("er".equals(name)) {
            addElement(Constant.Error);
        }
        // Ignore all other start tags.
    } // startElement

    /** Called by the SAXParser after parsing the end tag of an element.
     * @param uri currenly not documented; appears always to be ''.
     * @param localName The local name (without prefix), or the empty string if
     *                  Namespace processing is not being performed.  Currently
     *                  ignored.
     * @param name The qualified name (with prefix), or the empty string if
     *                  qualified names are not available.
     */
    public void endElement(String uri,
                             String localName,
                             String name)
    {
        try {
            if ("n".equals(name)) {
                // <n> number </n>
                String data = pcdata.toString().trim();
                Expr value;
                try {
                    // First try for zero
                    if (data.equals("0")) {
                        value = Constant.getInstance(0);
                    // Next check for a hex constant.
                    } else if (data.charAt(0) == '0'
                        && (data.charAt(1) == 'x' || data.charAt(1) == 'X'))
                    {
                        value = Constant.getInstance(
                                Integer.parseInt(data.substring(2), 16));
                    } else {
                        int i;
                        // Next check for an integer (all digits)
                        for (i = 0; i < data.length(); i++) {
                            if (!Character.isDigit(data.charAt(i))) {
                                break;
                            }
                        }
                        if (i == data.length()) {
                            // all digits
                            if (data.charAt(0) == '0') {
                                // octal integer
                                value = Constant.getInstance(
                                    Integer.parseInt(data.substring(1), 8));
                            } else {
                                // decimal integer
                                value = Constant.getInstance(
                                    Integer.parseInt(data));
                            }
                        // Finally, just assume floating point (thows an
                        // exception if the assumption is wrong).
                        } else {
                            value = Constant.getInstance(
                                Double.parseDouble(data));
                        }
                    }
                    addElement(value);
                } catch (NumberFormatException e){
                    addElement(
                        errorConstant(
                            "ill-formed number '" + data + "'", locator));
                }
                pcdata.setLength(0);
            } else if ("s".equals(name)) {
                addElement(Constant.getInstance(pcdata.toString()));
                pcdata.setLength(0);
            } else if ("t".equals(name)) {
                String data = pcdata.toString().trim();
                parsePos.setIndex(0);
                Date d = dateAndTimeFormat.parse(data, parsePos);
                if (d == null) {
                    d = dateFormat.parse(data, parsePos);
                }
                if (d == null) {
                    addElement(
                        errorConstant(
                            "ill-formed date '" + data + "'", locator));
                } else {
                    addElement(Constant.getInstance(d));
                }
                pcdata.setLength(0);
            } else if ("e".equals(name)) {
                String data = pcdata.toString().trim();
                if (caParser == null) {
                    caParser = new ClassAdParser(data);
                    caParser.setVerbosity(0);
                } else {
                    caParser.reset(data);
//Expr.db("*** verb " + caParser.setVerbosity(0));
                }
                Expr value = caParser.parse();
                if (value == null) {
                    value = errorConstant("invalid expression '" + data + "'",
                                locator);
                }
                addElement(value);
                pcdata.setLength(0);
            } else if ("ca".equals(name) ||"l".equals(name)) {
                Expr elt = (Expr) stack.pop();
                addElement(elt);
            }
            // The remaining tags have no content, so they are all handled by
            // startElement().
        } catch (Throwable t) {
            // FIXME
            // The current implementation of org.apache.crimson.parser.Parser2
            // apparently wraps a call to this method inside an promiscuous
            // catch clause that somehow loses the stack trace.  You get
            // a trace that ends with Parser2.parseInternal(Parser2.java:524)
            // losing all info about the real point where the error occurred.
            t.printStackTrace();
            System.exit(1);
        }
    } // endElement

    // ******** Private helper methods ***********

    /** For debugging, print the current stack. */
    private void printStack() {
        for (int i = 0; i < stack.size(); i++) {
            Expr e = (Expr)stack.elementAt(i);
            Expr.db("  " + e.toString(Expr.SHOW_ERROR_DETAIL));
        }
    } // printStack

    /** Helper function to dispose of a completed element.
     * If the stack is empty (this is a "top-level" element), just push it
     * onto the stack.
     * If this element is
     * the body of a &lt;a&gt; tag (as indicated by attrName != null), add it
     * to the containing RecordExpr.  Otherwise, it must a member of a list,
     * (or a &lt;classads&gt; element, which amounts to the same thing,)
     * so add it to the containing list.
     * @param e the parsed element to add.
     */
    private final void addElement(Expr e) {
        if (stack.empty()) {
            stack.push(e);
        } else {
            Object top = stack.peek();
            if (attrName == null) {
                ((ListExpr) top).add(e);
            } else {
                ((RecordExpr) top).insertAttribute(attrName, e);
                attrName = null;
            }
        }
    } // addElement

    /** Helper function for error() and warning().  Print an appropriate
     * message.
     * @param msg "Error" or "Warning"
     * @param e the exception associated with the error or warning.
     */
    private void printMessage(String msg, SAXParseException e) {
        if (verbosity == 0) {
            return;
        }
        System.err.print(msg + " at line " + e.getLineNumber());
        int col = e.getColumnNumber();
        if (col > 0) {
            System.err.print(" col " + col);
        }
        String id = e.getPublicId();
        if (id == null) {
            id = e.getSystemId();
        }
        if (verbosity > 1) {
            System.err.println(" of " + id + ":\n    " + e);
        } else {
            System.err.println(e.getMessage());
        }
    } // printMessage(String,SAXParseException)

    /** Format a parsing error as an ERROR constant.
     * @param msg information about the error.
     * @param locator source for line-number information.
     * @return an ERROR constant.
     */
    private static Expr errorConstant(String msg, Locator locator) {
        StringBuffer sb = new StringBuffer(msg);
        if (locator != null) {
            sb.append(" at line ")
                .append(locator.getLineNumber());
            int col = locator.getColumnNumber();
            if (col > 0) {
                System.err.print(" col " + col);
            }
        }
        return Constant.error(sb.toString());
    } // errorConstant(String,Locator)
} // ClassAdSAXHandler
