/** $Header: /p/condor/repository/CONDOR_SRC/src_java/condor/classad/CondExpr.java,v 1.11 2002/07/20 21:21:28 roy Exp $ */

/*--------------------------------------------------------------------
 *
 * Condor Java ClassAd library
 * Copyright (C) 1990-2001, CONDOR Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and Marvin Solomon
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General
 * Public License as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *--------------------------------------------------------------------*/

package condor.classad;

/**
 * A conditional expression.
 * An internal (non-leaf) node of an expression tree representing the ternary
 * operator <i>cond</i> ? <cond>expr1</cond> : <cond>expr2</cond>.
 * @see Expr
 * @author <a href="mailto:solomon@cs.wisc.edu">Marvin Solomon</a>
 * @version 1.2
 */
public class CondExpr extends Expr {
    private static String VERSION = "$Id: CondExpr.java,v 1.11 2002/07/20 21:21:28 roy Exp $";

    /** The condition (first component of the expression) */
    public final Expr ec;

    /** The true part (second component of the expression) */
    public final Expr et;

    /** The false part (third component of the expression) */
    public final Expr ef;

    /** Construct a conditional expression from its three components.
     * @param ec the condition (first component of the expression)
     * @param et the true part (second component of the expression)
     * @param ef the false part (third component of the expression)
     */
    public CondExpr(Expr ec, Expr et, Expr ef) {
        super(COND);
        this.ec = ec; this.et = et; this.ef = ef;
    } // CondExpr(Expr,Expr,Expr)

    /** Convert this Expr to a string.
     * The exact format of the result is controled by the static fields
     * <code>Expr.formatFlags</code> and <code>Expr.formatIndent</code>.
     * The result may be multiple lines long (it may include embedded newlines
     * and tabs), but it never ends with a newline.
     * @return a string representation of the this Context
     * @see Expr#formatFlags
     * @see Expr#formatIndent
     */
    public String toString() {
        String
            ec1 = ec.toString(formatFlags & ~XML | XML_CONTENT),
            et1 = et.toString(formatFlags & ~XML | XML_CONTENT),
            ef1 = ef.toString(formatFlags & ~XML | XML_CONTENT);
        String prefix = ((formatFlags & XML) != 0) ? "<e>" : "";
        String suffix = ((formatFlags & XML) != 0) ? "</e>" : "";
        if ((formatFlags & MINIMAL_PARENTHESES) == 0) {
            // Minimal number of parentheses
            return prefix + "(" + ec1 + "?" + et1 + ":" + ef1 + ")" + suffix;
        } else {
            return
                prefix
                + (ec.prec() <= prec() ?  "(" + ec1 + ")" : ec1)
                + "?" + et1
                + ":" + ef1
                + suffix;
        }
    } // toString()

    /** The precedence of the operator in this expression node.
     * Used to print expressions without superfluous parentheses.
     * @return the precendence of this node.
     * @see Expr#MINIMAL_PARENTHESES
     */
    public int prec() {
        return -1;
    } // prec()

    /** Evalutate this expression.
     * @param ctx an evaluation context for evaluating this expression
     * @return a Constant, ListExpr, or RecordExpr representing the value.
     * @see Expr#eval(condor.classad.Context)
     */
    protected Expr reEval(Context ctx) {
        Expr ecv = ec.eval(ctx);

        if (ecv == Constant.Undef || ecv == Constant.Error)
            return ecv;
        if (ecv.type != BOOLEAN)
            return
                Constant.error(
                    "type " + ecv.typeName()
                    + " found where boolean expected"
                    + " in conditional expression");
        return ((Constant)ecv).toBoolean() ? et.eval(ctx) : ef.eval(ctx);
    } // reEval(Context)
} // CondExpr
