/** $Header: /p/condor/repository/CONDOR_SRC/src_java/condor/classad/Constant.java,v 1.11 2002/07/20 21:21:28 roy Exp $ */

/*--------------------------------------------------------------------
 *
 * Condor Java ClassAd library
 * Copyright (C) 1990-2001, CONDOR Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and Marvin Solomon
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General
 * Public License as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *--------------------------------------------------------------------*/

package condor.classad;

import java.util.*;


/**
 * A constant (Integer, Real, etc.).  The constant is wrapped so that it
 * may be used as a node in an expression tree.  To conserve space, there
 * is at most one Constant instance with any given value.
 * @see Expr
 * @author <a href="mailto:solomon@cs.wisc.edu">Marvin Solomon</a>
 * @version 1.2
 */
public class Constant extends Expr {
    private static String VERSION = "$Id: Constant.java,v 1.11 2002/07/20 21:21:28 roy Exp $";

    /** The actual value of this Constant.  Usage depends on type:
     * <dl>
     * <dt>UNDEFINED     <dd>A String error message.
     * <dt>ERROR         <dd>A String error message.
     * <dt>BOOLEAN       <dd>null (there are only two Boolean Constants).
     * <dt>INTEGER       <dd>An Integer value.
     * <dt>REAL          <dd>A Double value.
     * <dt>STRING        <dd>A String value.
     * <dt>ABSOLUTE_TIME <dd>A java.util.Date value.
     * <dt>RELATIVE_TIME <dd>A Long value (number of milliseconds).
     * </dl>
     */
    public final Object value;

    /** The set of all extant instances of Constant, indexed by value.
     * This Map is used to save storage overhead by ensuring that there
     * is at most one Constant with the same value.
     */
    private static Map instanceMap = new HashMap();

    /** Construct an arbitrary constant.
     * Note that this is the only constructor and it is private.
     * Clients create new instances by calling the factory methods
     * newInstance, error, undefined, or bool.
     * @param value the value of this constant.
     * @param type the type of this constant.
     */
    private Constant(Object value, int type) {
        super(type);
        this.value = value;
    } // Constant(Object,int)

    /** A default undefined constant */
    public static final Constant Undef
        = new Constant("", UNDEFINED);

    /** A default error constant */
    public static final Constant Error
        = new Constant("", ERROR);

    /** The unique boolean constant true */
    public static final Constant True
        = new Constant(null, BOOLEAN);

    /** The unique boolean Constant false */
    public static final Constant False
        = new Constant(null, BOOLEAN);

    /** An INTEGER Constant with the value 0, to expedite common cases. */
    static private final Constant INT_ZERO =
        new Constant(new Integer(0), INTEGER);

    /** An INTEGER Constant with the value 1, to expedite common cases. */
    static private final Constant INT_ONE =
        new Constant(new Integer(1), INTEGER);

    /** A REAL Constant with the value 0, to expedite common cases. */
    static private final Constant REAL_ZERO =
        new Constant(new Double(0.0), REAL);

    /** A REAL Constant with the value 1, to expedite common cases. */
    static private final Constant REAL_ONE =
        new Constant(new Double(1.0), REAL);

    /** A STRING Constant with the value "", to expedite common cases. */
    static private final Constant EMPTY_STRING = new Constant("", STRING);

    /** Returns the unique constant of type "type" and value "val", creating
     * one if necessary.
     * @param val the value.
     * @param type the type.
     * @return a new constant.
     */
    private static final Constant getInstance(Object val, int type) {
        Constant result = (Constant) instanceMap.get(val);
        if (result == null) {
            result = new Constant(val, type);
            instanceMap.put(val, result);
        }
        return result;
    } // getInsteance(Object,type)

    /** Returns the unique integer constant with value "value", creating one if
     * necessary.
     * @param value the value of the Constant.
     */
    public static final Constant getInstance(int value) {
        if (value == 0) {
            return INT_ZERO;
        }
        if (value == 1) {
            return INT_ONE;
        }
        return getInstance(new Integer(value), INTEGER);
    } // getInstance(int)

    /** Returns the unique real constant with value "value", creating one if
     * necessary.
     * @param value the value of the Constant.
     */
    public static final Constant getInstance(double value) {
        if (value == 0.0) {
            return REAL_ZERO;
        }
        if (value == 1.0) {
            return REAL_ONE;
        }
        return getInstance(new Double(value), REAL);
    } // getInstance(double)

    /** Returns the unique String constant with value "value", creating one if
     * necessary.
     * @param value the value of the Constant.
     */
    public static final Constant getInstance(String value) {
        if (value == null) {
            return null;
        }
        if (value.length() == 0) {
            return EMPTY_STRING;
        }
        return getInstance(value, STRING);
    } // getInstance(String)

    /** Create a String Constant from a character array, processing backslash
     * escapes.
     * @param buf the array of characters.
     * @param start the index preceding the first character to be used.
     * @param stop the index following the last characater to be used.
     */
    public static final Constant getInstance(char[] buf, int start, int stop) {
        return getInstance(unquoteString(buf, start, stop));
    } // getInstance(char[],int,int)

    /** Create an absolute time Constant.
     * @param value the value of the Constant.
     */
    public static final Constant getInstance(Date value) {
        return getInstance(value, ABSOLUTE_TIME);
    } // getInstance(Date)

    /** Returns the unique time constant with value "value", creating one if
     * necessary.
     * @param value the value of the constant in milliseconds
     */
    public static final Constant getInstance(long value) {
        return getInstance(new Long(value), RELATIVE_TIME);
    } // getInstance(long)

    /** Convert to a string
     * @return the value of this Constant, converted to a string.
     */
    public String toString() {
        boolean xml = (formatFlags & XML) != 0;
        switch (type) {
            case UNDEFINED:
                if (xml) {
                    return "<un/>";
                } else if ((formatFlags & SHOW_ERROR_DETAIL) != 0) {
                    return "UNDEFINED(" + value + ")";
                } else {
                    return "UNDEFINED";
                }
            case ERROR:
                if (xml) {
                    return "<er/>";
                } else if ((formatFlags & SHOW_ERROR_DETAIL) != 0) {
                    return "ERROR(" + value + ")";
                } else {
                    return "ERROR";
                }
            case BOOLEAN:
                if (xml) {
                    return
                        this == True ? "<b v=\"t\"/>" : "<b v=\"f\"/>";
                } else {
                    return this == True ? "true" : "false";
                }
            case INTEGER:
            case REAL:
                return
                    xml
                        ? ("<n>" + value.toString() + "</n>")
                        : value.toString();
            case STRING:
                if (xml) {
                    return "<s>" + quoteStringXML((String) value) + "</s>";
                } else if ((formatFlags & XML_CONTENT) != 0) {
                    return '"' + quoteStringXML((String) value) + '"';
                } else if ((formatFlags & NO_ESCAPE_STRINGS) != 0) {
                    return (String) value;
                } else {
                    return quoteString((String) value);
                }
            case ABSOLUTE_TIME:
                return
                    xml ? ("<t>" + value.toString() + "</t>")
                    : value.toString(); // TODO: should be surrounded by '' ?
            case RELATIVE_TIME:
            {
                long v = ((Long) value).longValue();
                return
                    xml ? ("<t>" + rel2str(v) + "</t>")
                    : rel2str(v); // TODO: should be surrounded by '' ?
            }
            default: throw new RuntimeException("unknown type");
        }
    } // toString()

    /** Convernience function: Convert to a boolean.
     * Note that unlike realValue, intValue, etc, this method never
     * throws an exception.
     * @return true if this constant is a BOOLEAN with the value true;
     * false in all other cases.
     */
    public boolean toBoolean() {
        if (type != BOOLEAN)
            return false;
        return this == True;
    } // toBoolean()

    /** Convenience function:  Create an ERROR constant containing a particular
     * message.
     * @param msg the message.
     * @return the Constant.
     */
    public static Constant error(String msg) {
        return new Constant(msg, ERROR);
    } // error(String)

    /** Convenience function:  Create an UNDEFINED constant containing a
     * particular message.
     * @param msg the message.
     * @return the Constant.
     */
    public static Constant undefined(String msg) {
        return new Constant(msg, UNDEFINED);
    } // undefined(String)

    /** Convenience function:  Convert a boolean to Constant.True or
     * Constant.False.
     * @param b the boolean value.
     * @return the Constant.
     */
    public static Constant bool(boolean b) {
        return b ? True : False; 
    } // bool(boolean)

    // Accessors.  Each method coverts the value to the indicated type if
    // possible.  Otherwise, it throws ArithmeticException.

    /** Get the integer value of this Constant if possible.
     * @exception ArithmeticException if this Constant is not an integer.
     * @return the integer value of this IntegerConstant.
     */
    public int intValue() throws ArithmeticException {
        if (type != INTEGER)
            throw new ArithmeticException(typeName() + " " + this
                + " in integer context");
        return ((Integer)value).intValue();
    } // intValue()

    /** Get the double floating point value of this Constant if possible.
     * @exception ArithmeticException if this Constant is not an instance of
     * IntegerConstant or RealConstant.
     * @return the real value of this RealConstant or the value of this
     * IntegerConstant converted to double.
     */
    public double realValue() throws ArithmeticException {
        switch (type) {
            case REAL:
                return ((Double)value).doubleValue();
            case INTEGER:
                return ((Integer)value).doubleValue();
        }
        throw new ArithmeticException(typeName() + " " + this
            + " in real context");
    } // realValue()

    /** Get the string value of this Constant if possible.
     * @exception ArithmeticException if this Constant is not an instance of
     * StringConstant.
     * @return the string value of this StringConstant.
     */
    public String stringValue() throws ArithmeticException {
        if (type != STRING)
            throw new ArithmeticException(typeName() + " " + this
                + " in string context");
        return (String)value;
    } // stringValue()

    /** Convert a time value to milliseconds.
     * @exception ArithmeticException if this Constant is not a time value.
     * @return the number of milliseconds (since 1/1/70 for ABSOLUTE_TIME
     * values).
     */
    public long milliseconds() throws ArithmeticException {
        if (type == ABSOLUTE_TIME)
            return ((Date)value).getTime();
        if (type == RELATIVE_TIME )
            return ((Long)value).longValue();
        throw new ArithmeticException(typeName() + " " + this
            + " in integer context");
    } // milliseconds()

    /** The precedence of this expression node (MAXPREC).
     * Used to print expressions without superfluous parentheses.
     * @return the precendence of this node.
     * @see Expr#MINIMAL_PARENTHESES
     */
    public int prec() { return MAXPREC+1; }

    /** Generate a string from an array of characters, processing backslash
     * escapes.  The the characters buf[start+1] ... buf[stop-1] are
     * converted to a String, after replacing backslash escapes, such as \n by
     * their values.
     * @param buf the array of characters.
     * @param start the index preceding the first character to be used.
     * @param stop the index following the last characater to be used.
     * @return the converted String.
     */
    private static final String unquoteString(char[] buf, int start, int stop) {
        char[] result = new char[stop - start - 1];
        int i = start+1;
        int j = 0;
        while (i < stop) {
            char c = buf[i++];
            if (c == '\\') {
                c = buf[i++];
                int val = Character.digit(c,8);
                if (val >= 0) {
                    int maxDigits = (val < '4') ? 2 : 1;
                    for (int k=0; k<maxDigits; k++) {
                        c = buf[i];
                        int n = Character.digit(c, 8);
                        if (n < 0)
                            break;
                        val = (val<<3) + n;
                        i++;
                    }
                    c = (char)val;
                }
                else switch (c) {
                    case 'n': c = '\n'; break;
                    case 't': c = '\t'; break;
                    case 'b': c = '\b'; break;
                    case 'r': c = '\r'; break;
                    case 'f': c = '\f'; break;
                    default: break;
                }
            }
            result[j++] = c;
        }
        return new String(result, 0, j);
    } // unquoteString(char[],int,int)

    /** Quote a string, replacing non-printable values by backslash escapes.
     * Note that Unicode is not supported: We _assume_ the value of
     * every character is in the range 0..0377.
     * @param s the string to be quoted.
     * @return a quoted and escaped version of the string.
     */
    public final static String quoteString(String s) {
        // First compute the length of the result
        int len = 0;
        for (int i=0; i<s.length(); i++) {
            int c = s.charAt(i) & 0xff;
            switch (c) {
                case 'n': case 't': case 'b': case 'r': case 'f':
                case '\\': case '\"':
                    len += 2;
                    break;
                default:
                    if (c < 040 || c > 0176) len += 4; // \ddd
                    else len++;
            }
        }
        char [] result = new char[len+2];
        int j = 0;
        result[j++] = '"';
        for (int i=0; i<s.length(); i++) {
            int c = s.charAt(i) & 0xff;
            switch (c) {
                case '\n': result[j++] = '\\'; result[j++] = 'n'; break;
                case '\t': result[j++] = '\\'; result[j++] = 't'; break;
                case '\b': result[j++] = '\\'; result[j++] = 'b'; break;
                case '\r': result[j++] = '\\'; result[j++] = 'r'; break;
                case '\f': result[j++] = '\\'; result[j++] = 'f'; break;
                case '\\': result[j++] = '\\'; result[j++] = '\\'; break;
                case '\"': result[j++] = '\\'; result[j++] = '\"'; break;
                default:
                    if (c < 040 || c > 0176) {
                        result[j] = '\\';
                        for (int k=3; k>=1; --k) {
                            result[j+k] = (char)('0' + (c & 7));
                            c >>= 3;
                        }
                        j += 4;
                    }
                    else result[j++] = (char)c;
                    break;
            }
        }
        result[j++] = '"';
        return new String(result, 0, j);
    } // quoteString(String)

    /** Fix a string to use XML exscape sequences for non-printable characters
     * and the characters '&lt;', '&amp;', and '&gt;'.
     * Note that Unicode is not supported: We _assume_ the value of
     * every character is in the range 0..0377.
     * @param s the string to be quoted.
     * @return an escaped version of the string.
     */
    private final static String quoteStringXML(String s) {
        StringBuffer result = new StringBuffer(s);
        for (int i = 0; i < result.length(); ) {
            int c = result.charAt(i) & 0xff;
            switch (c) {
            case '<':
                result.replace(i, i+1, "&lt;");
                i += 4;
                break;
            case '>':
                result.replace(i, i+1, "&gt;");
                i += 4;
                break;
            case '&':
                result.replace(i, i+1, "&amp;");
                i += 5;
                break;
            case '"':
                result.replace(i, i+1, "&quot;");
                i += 6;
                break;
            default:
                if (c < 040 || c > 0176) {
                    result.replace(i, i+1,
                        "&#x" + Integer.toHexString((c >> 4) & 0xf)
                              + Integer.toHexString(c & 0xf)
                              + ";");
                    i += 6; // &#xab;
                } else {
                    i++;
                }
            }
        }
        return result.toString();
    } // quoteStringXML(String)

    /** Helper table for rel2str. */
    private static int[] threshold = {
        24*60*60*1000,
        60*60*1000,
        60*1000,
        1000
    };

    /** Helper table for rel2str. */
    private static final char[] seperator = { '+', ':', ':', '.' };

    /** Helper method for rel2str.
     * Convert n to a decimal string, optionally with leading zeros.
     * @param buf the place to put the result.
     * @param n the number to be converted (must be non-negative).
     * @param a power of 10 indicating the minimum number of digits in the
     *    result.  For example, if m==100, the result will have at least three
     *    digits.
     */
    private static final void formatDecimal(StringBuffer buf, int n, int m) {
        buf.append((int)(n/m));
        while (m > 1) {
            n %= m;
            m /= 10;
            buf.append((int)(n/m));
        }
    } // formatDecimal(StringBuffer,int,int)

    /** Format a relative time value as a readable String.
     * The result is a string of of the form days+hh:mm:ss.mmm, where
     * leading components are omitted if they are zero.
     * For example,
     * <ul>
     * <li> ten milliseconds becomes ".010".
     * <li> twelve minutes and thirty-four seconds becomes "12:34.000"
     * </ul>
     * @param t the number to be converted, in milliseconds.
     * @return the resulting String.
     */
    private static final String rel2str(long t) {
        StringBuffer buf = new StringBuffer();
        int i = 0;
        if (t < 1000) {
            buf.append("0.");
        }
        else {
            while (t <= threshold[i])
                i++;
            formatDecimal(buf, (int)(t/threshold[i]), 1);
            for (;;) {
                buf.append(seperator[i]);
                t %= threshold[i];
                i++;
                if (i >= threshold.length)
                    break;
                formatDecimal(buf, (int)(t/threshold[i]), 10);
            }
        }
        formatDecimal(buf, (int)t, 100);
        return buf.toString();
    } // rel2str(long)
} // Constant
