package condor.classad;

/*--------------------------------------------------------------------
 *
 * Condor Java ClassAd library
 * Copyright (C) 1990-2001, CONDOR Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and Marvin Solomon
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General
 * Public License as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *--------------------------------------------------------------------*/

import java.util.Hashtable;
import java.util.Enumeration;

/** $Header: /p/condor/repository/CONDOR_SRC/src_java/condor/classad/Context.java,v 1.7 2002/07/20 21:21:28 roy Exp $ */

/**
 * A context for evaluating ClassAd expressions.  It consists of a pointer to
 * the outermost classad of the ``top-level'' expression being evaluated and
 * two hash tables.
 * <p>The first table maps subexpressions to the results of evaluating them.  It
 * serves not only to ``memoize'' computed results as a shortcut to speed up
 * evaluation but also (perhaps more importantly) to detect circular references.
 * <p>The second table maps classad expressions to their parent scopes.
 * The parent scope of a classad is normally the classad that immediately
 * encloses statically, but may be any classad expression in special cases.
 * In particular, this flexibility is used by ClassAd.bind to create a Context
 * in which each of two RecordExprs is the parent of the other.
 * <p>Eventually, other transient state may be added to this class to allow
 * iteration through a set of results, etc.
 * @see ClassAd#bind(condor.classad.RecordExpr,
 * condor.classad.RecordExpr)
 */
public class Context {
    private static String VERSION = "$Id: Context.java,v 1.7 2002/07/20 21:21:28 roy Exp $";

    /** The outermost classad of the expression being evaluated.
     * There must always be a unique outermost classad when an expression is
     * evaluated.  This field is used to
     * evaluate absolute attribute references of the form ``.name''.
     */
    public final RecordExpr root;

    /** A mapping from subexpressions to the results of evaluating them.
     */
    private Hashtable valueMap = new Hashtable();

    /** A mapping from RecordExpr subexpressions to enclosing RecordExpr.
     */
    private Hashtable parentMap = new Hashtable();

    /** A special flag used to mark a hashtable entry as "in progress" to
     * detect loops.
     */
    static private final Expr inProgress
        = Constant.undefined("circular attribute reference");

    /** Create a new context.
     * @param root the outermost classad in the expression being evaluated.
     */
    public Context(RecordExpr root) {
        this.root = root;
    }

    /** Look for a cached evaluation.
     * If none is found, return null, but mark the transalation as ``in
     * progress''.
     * @param expr the Expr node.
     * @return the translation, or null if none found
     */
    public Expr getValue(Expr expr) {
        Object o = valueMap.get(expr);
        if (o == null) {
            valueMap.put(expr, inProgress);
            return null;
        }
        return (Expr)o;
    }

    /** Record the computed value of a given node.
     * @param expr the given Expr node.
     * @param value the corresponding value.
     * @return this Context
     */
    public Context setValue(Expr expr, Expr value) {
        valueMap.put(expr, value);
        return this;
    }

    /** Look for a cached parent binding.
     * @param expr a RecordExpr node.
     * @return the translation, or null if none found
     */
    public RecordExpr getParent(RecordExpr expr) {
        Object o = parentMap.get(expr);
        if (o == null)
            return null;
        return (RecordExpr)o;
    }
    
    /** Record the parent RecordExpr of a given node
     * @param expr the given RecordExpr node.
     * @param value the enclosing RecordExpr.
     * @return this Context.
     */
    public Context setParent(RecordExpr expr, RecordExpr value) {
        parentMap.put(expr, value);
        return this;
    }

    /** Convert this Context to a string, for debugging purposes.
     * @return a string representation of the this Context
     */
    public String toString() {
        String res = "Context\n" + Expr.formatIndent
            + "   root = " + root.mark(true)
            + "\n" + Expr.formatIndent + "   valueMap={";
        int n = 0;
        for (Enumeration k = valueMap.keys(); k.hasMoreElements();) {
            Expr x = (Expr)k.nextElement();
            Expr y = (Expr)valueMap.get(x);
            if (n++ > 0) res += ", ";
            res += x.mark(true) + "->";
            if (y == null || y instanceof Constant) res += y;
            else res += y.mark(true);
        }
        res += "}\n" + Expr.formatIndent + "   parentMap={";
        n = 0;
        for (Enumeration k = parentMap.keys(); k.hasMoreElements();) {
            Expr x = (Expr)k.nextElement();
            Expr y = (Expr)parentMap.get(x);
            if (n++ > 0) res += ", ";
            res += x.mark(true) + "->";
            if (y == null || y instanceof Constant) res += y;
            else res += y.mark(true);
        }
        return res + "}";
    }
} // Context
