/** $Header: /p/condor/repository/CONDOR_SRC/src_java/condor/classad/Expr.java,v 1.16 2002/07/20 21:21:28 roy Exp $ */

/*--------------------------------------------------------------------
 *
 * Condor Java ClassAd library
 * Copyright (C) 1990-2001, CONDOR Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and Marvin Solomon
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General
 * Public License as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *--------------------------------------------------------------------*/

package condor.classad;
import java.text.Collator;
import java.text.CollationKey;

/**
 * A node of an expression tree.  An Expr is one of
 * <dl>
 *   <dt>{@link RecordExpr}
 *       <dd>a "classad" of the form <code>[name = expr, ... ]</code>,
 *   <dt>{@link SelectExpr}
 *       <dd>a selection from a record as in <code>a.b</code>,
 *   <dt>{@link ListExpr}
 *       <dd>a list of the form <code>{ expr1, expr2, ... }</code>,
 *   <dt>{@link SubscriptExpr}
 *       <dd>a selection from a list as in <code>a[i]</code>,
 *   <dt>{@link CondExpr}
 *       <dd>a ternary conditional expression as in <code>a ? b : c</code>,
 *   <dt>{@link Op}
 *       <dd>a binary or unary operator applied to operand(s),
 *   <dt>{@link FuncCall}
 *       <dd>a call to a function, with a list of argument expressions,
 *   <dt>{@link AttrRef}
 *       <dd>a reference to an attribute of a ClassAd, of the form
 *                  <code>[.]identifier[.identifier]</code>, or
 *   <dt>{@link Constant}
 *       <dd>a constant (literal) value
 * </dl>
 *
 * @see RecordExpr
 * @see SelectExpr
 * @see ListExpr
 * @see SubscriptExpr
 * @see CondExpr
 * @see Op
 * @see FuncCall
 * @see AttrRef
 * @see Constant
 * @author <a href="mailto:solomon@cs.wisc.edu">Marvin Solomon</a>
 * @version 1.2
 */
public abstract class Expr {
    private static String VERSION = "$Id: Expr.java,v 1.16 2002/07/20 21:21:28 roy Exp $";

    /** The "type" of this expression.  For constants, the type is one of
     * UNDEFINED,
     * ERROR,
     * BOOLEAN,
     * INTEGER,
     * REAL,
     * STRING,
     * ABSOLUTE_TIME, or
     * RELATIVE_TIME.
     * <p>
     * For non-atomic values, the type is one of 
     * COND,
     * OP,
     * CALL,
     * SELECTION,
     * SUBSCRIPT,
     * ATTRIBUTE,
     * LIST, or
     * RECORD.
     * Note that these values are all less than zero.
     */
    public final int type;

    /** Create an Expr node of a given type.  Used in constructors of
     * subclasses.
     */
    protected Expr(int type) {
        this.type = type;
    } // Expr(int)

    /** A type flag indicating that this is a conditional expression */
    static public final int COND = -8;

    /** A type flag indicating that this is a unary or binary operator */
    static public final int OP = -7;

    /** A type flag indicating that this is a function call */
    static public final int CALL = -6;

    /** A type flag indicating that this is a selection (a.b) */
    static public final int SELECTION = -5;

    /** A type flag indicating that this is a selection (a.b) */
    static public final int SUBSCRIPT = -4;

    /** A type flag indicating that this is an attribute reference */
    static public final int ATTRIBUTE = -3;

    /** A type flag indicating that this is a list */
    static public final int LIST = -2;

    /** A type flag indicating that this is a record (classad) */
    static public final int RECORD = -1;

    /** A type flag indicating that this is an undefined value */
    static public final int UNDEFINED = 0;

    /** A type flag indicating that this is an error value */
    static public final int ERROR = 1;

    /** A type flag indicating that this is a boolean value */
    static public final int BOOLEAN = 2;

    /** A type flag indicating that this is an integer value */
    static public final int INTEGER = 3;

    /** A type flag indicating that this is a real value */
    static public final int REAL = 4;

    /** A type flag indicating that this is a string value */
    static public final int STRING = 5;

    /** A type flag indicating that this is an absolute time value */
    static public final int ABSOLUTE_TIME = 6;

    /** A type flag indicating that this is a relative time value */
    static public final int RELATIVE_TIME = 7;

    /** Mapping of type codes to descriptive strings, used by typeName(). */
    static private final String[] tNames = {
        "conditional expression",
        "operator expression",
        "function call",
        "selection",
        "subscripted expression",
        "attribute reference",
        "list",
        "record",
        "undefined",
        "error",
        "boolean",
        "integer",
        "real",
        "string",
        "timestamp",
        "time interval"
    }; // tNames

    /** Determine the type of this expression (for printing messages). 
     * @return a string identifing the type of this constant.
     */
    protected String typeName() {
        return tNames[type - COND];
    } // typeName()

    /** A marker to distinguish different instances of sub-expressions in
     * a debugging print-out.
     * This is a temporary debugging feature that may be removed from future
     * revisions of this package.
     */
    public final int instance = nextMarker++;

    /** A counter for generating values for the instance field.
     * This is a temporary debugging feature that may be removed from future
     * revisions of this package.
     */
    private static int nextMarker = 1;

    /** Conditionally return an indication of the unique instance represented
     * by this expression.
     * This is a temporary debugging feature that may be removed from future
     * revisions of this package.
     * @param on indicates whether to indicate the particular instance.
     * @return if <code>on</code> is true, a string of the form
     * &lt;nn&gt;, where nn is a decimal integer, otherwise an empty string.
     */
    protected String mark(boolean on) {
        return on ? ("<"+instance+">") : "";
    } // mark(boolean)

    /** Conditionally return an indication of the unique instance represented
     * by this expression.
     * This is a temporary debugging feature that may be removed from future
     * revisions of this package.
     * @return if <code>Expr.formatFlags</code> includes the flag
     * <code>Expr.SHOW_INSTANCES</code>, a string of the form
     * &lt;nn&gt;, where nn is a decimal integer, otherwise an empty string.
     */
    protected String mark() {
        return mark((formatFlags & SHOW_INSTANCES)!=0);
    } // mark()

    /** Return an indication of the unique instance of the argument.
     * This is a temporary debugging feature that may be removed from future
     * revisions of this package.
     * @param expr an expression instance or <code>null</code>
     * @return the string &lt;&gt; if <code>expr==null</code>, otherwise 
     * <code>expr.mark()</code>.
     * @see #mark()
     */
    protected static String mark(Expr expr) {
        return
            (formatFlags & SHOW_INSTANCES)==0
                ? ""
            : expr==null
                ? "<>"
            : expr.mark();
    } // mark(Expr)

    /** A Collator for case-insensitive comparison of hash keys. */
    public static final Collator caseIgnore;

    /** A static initialization block to intialize caseIgnore.
     * @see #caseIgnore
     */
    static {
        caseIgnore = Collator.getInstance();
        caseIgnore.setStrength(Collator.PRIMARY);
    } // static

    /** Evaluate this Expr.
     * The model is ``evalutate until normal form''.  A normal form expression
     * (currently a Constant, ListExpr, or RecordExpr) evaluates to itself.
     * Any other expression is evaluated by evaluating its operand(s) as
     * necessary and then applying its top-level operator to the results.
     * If the types of the evaluated operands are wrong, for example if the
     * top-level operator is + and an operand evaluates to a RecordExpr, the
     * result is an ErrorConstant, which is in normal form.
     * If the result of applying the operator is not in normal form, it is
     * re-evaluated.
     * <p>The Context records a mapping from non-normal-form expressions
     * to their values.  It speeds up evaluation and guarentees termination.
     * <p><b>Theorem</b>:
     * Evaluation is linear-time in the size of the original Expr.
     * <p><b>Proof</b>:
     * When evaluation of an expression is started, the Context maps it to an
     * UndefinedConstant.  If the evaluation process recursively requests the
     * value of the same expression, the result is that UndefinedConstant.
     * When the evaluation of a particular subexpression is complete, the
     * Context replaces the UndefinedConstant with a reference to the computed
     * value.  This value may be either a sub-expression of the original
     * expression or a newly created Expr value, but in any case, it will be in
     * normal form.  Thus a value found in the Context will never require
     * re-evaluation.
     * <p>In short, a top-level call to eval will never attempt to
     * evaluate any node more than once, and the total number of new nodes
     * created is bounded by the number of nodes in the original expression.
     * <b>QED</b>
     * @param ctx an evaluation context for evaluating this expression
     * @return a Constant, ListExpr, or RecordExpr representing the value.
     * @see Context
     */
    public Expr eval(Context ctx) {
        Expr result = this;
        if (dblevel >= 1) {
            db(formatIndent + ">>> start eval " + mark(true) + ": "
                + toString(READABLE|SHOW_INSTANCES));
        }
        if (dblevel>=2) {
            db(formatIndent + ctx);
            if (isNormalForm()) {
                db(formatIndent + " ----- normal form");
            }
        }
        for (Expr expr = this; !expr.isNormalForm(); expr = result) {
            // First look for a memoized result
            result = ctx.getValue(expr);
            if (result == null) {
                String saveIndent = formatIndent;
                if (dblevel>=2) {
                    db(formatIndent + "reEval");
                    formatIndent = "====" + formatIndent;
                }
                result = expr.reEval(ctx);
                ctx.setValue(expr, result);
                formatIndent = saveIndent;
                if (dblevel>=2) {
                    db(formatIndent + "reEval returns: "
                        + result.toString(READABLE|SHOW_INSTANCES));
                }
            }
            else {
                // In the case of a circular reference, the lookup will
                // return the special InProgress marker, which is an
                // UndefinedConstant, which is exactly the right thing to
                // return in this case.
                if (dblevel>=2) {
                    db(formatIndent + " found cached result: "
                        + result.toString(READABLE|SHOW_INSTANCES));
                }
            }
            expr = result;
        }
        ctx.setValue(this, result);
        if (dblevel>=2) db(formatIndent + ">>> end eval");
        return result;
    } // eval(Context)

    /** Indicate whether this is a normal-form expression.
     * @return true if this Expr is in normal form (a Constant, RecordExpr, 
     * or ListExpr)
     */
    protected final boolean isNormalForm() {
        return type >= LIST;
    } // isNormalForm()

    /** Indicate whether this is a constant.
     * @return true if this Expr is a constant.
     */
    protected final boolean isConstant() {
        return type >= 0;
    } // isConstant()

    /** Evalutate this expression.
     * This method is only called for expressions that are not in normal form.
     * @param ctx an evaluation context for evaluating this expression
     * @return a Constant, ListExpr, or RecordExpr representing the value.
     * @see #eval(condor.classad.Context)
     */
    protected Expr reEval(Context ctx) {
        return this;
    } // reEval(Context)

    /** A convenience function for creating a SelectExpression corresponding
     * to ``this.selector''
     * @param selector a selector.
     * @return a new SelectExpr corresponding to ``this.selector''.
     */
    public SelectExpr select(String selector) {
        return new SelectExpr(this, CiString.getInstance(selector));
    } // select(String)

    /** A standing indent for all expressions whose string representation spans
     * multiple lines.  This string is appended to the start of all lines but
     * the first.
     */
    static public String formatIndent = "";

    /** Flags to control the formatting of expressions by
     * <code>toString()</code>.  A value of zero creates a representation
     * that is unambiguous and easy to parse but not very readable.  Each bit
     * turns on a formatting feature that makes the string more readable, but
     * perhaps more verbose and/or ambiguous.
     * @see #NO_ESCAPE_STRINGS
     * @see #SHOW_ERROR_DETAIL
     * @see #MINIMAL_PARENTHESES
     * @see #MULTI_LINE_ADS
     * @see #MULTI_LINE_LISTS
     * @see #SHOW_INSTANCES
     * @see #XML
     * @see #XML_CONTENT
     * @see #BRIEF
     * @see #COMPACT
     * @see #READABLE
     */
    static protected int formatFlags = 0;

    /** An option flag for <code>toString()</code>.
     * Don't surround strings with quotes or escape special characters in the
     * string.
     * <p><strong>WARNING</strong>: Strings created with this flag cannot be
     * parsed to recover the expressions that created them.
     * @see #formatFlags
     */
    static public final int NO_ESCAPE_STRINGS = 1<<0;

    /** An option flag for <code>toString()</code>.
     * Print the ERROR value as ``ERROR(reason)'' rather than just ``ERROR''.
     * @see #formatFlags
     */
    static public final int SHOW_ERROR_DETAIL = 1<<1;

    /** An option flag for <code>toString()</code>.
     * Use only as many parentheses are are required to override the default
     * precendence of operators.
     * @see #formatFlags
     */
    static public final int MINIMAL_PARENTHESES = 1<<2;

    /** An option flag for <code>toString()</code>.
     * Display classads with one attribute per line.
     * @see #formatFlags
     */
    static public final int MULTI_LINE_ADS = 1<<3;

    /** An option flag for <code>toString()</code>.
     * Display lists with one member per line.
     * @see #formatFlags
     */
    static public final int MULTI_LINE_LISTS = 1<<4;

    /** An option flag for <code>toString()</code>.
     * Flag certain kinds of expressions with a marker indicating unique
     * instances.
     * This is a temporary debugging feature that may be removed from future
     * revisions of this package.
     * @see #formatFlags
     */
    static public final int SHOW_INSTANCES = 1<<5;

    /** An option flag for <code>toString()</code>.
     * Convert the expression to XML.
     * @see #formatFlags
     */
    static public final int XML = 1<<6;

    /** An option flag for <code>toString()</code>.
     * Internal flag inserted by the recursive traversal initiated by
     * ancestor.toString(Expr.XML).  Indicates that this part of XML content,
     * and therefore should escape any occurence of `&amp;' in the output
     * as `&amp;amp', `&lt;' as `&amp;lt;', etc.
     * @see #formatFlags
     */
    static public final int XML_CONTENT = 1<<7;

    /** A combination of option flags for <code>toString()</code>.
     * Produce a concise version (both MINIMAL_PARENTHESES and
     * NO_ESCAPE_STRINGS)
     * @see #formatFlags
     * @see #MINIMAL_PARENTHESES
     * @see #NO_ESCAPE_STRINGS
     */
    static public final int BRIEF = MINIMAL_PARENTHESES | NO_ESCAPE_STRINGS;

    /** A combination of option flags for <code>toString()</code>.
     * Produce a compact but parsable version (currently, the same as
     * MINIMAL_PARENTHESES).
     * @see #formatFlags
     * @see #MINIMAL_PARENTHESES
     */
    static public final int COMPACT = MINIMAL_PARENTHESES;


    /** A combination of option flags for <code>toString()</code>.
     * Produce a maximally-readable version (all flags other than
     * SHOW_INSTANCES).
     * @see #formatFlags
     * @see #SHOW_INSTANCES
     */
    static public final int READABLE =
                                NO_ESCAPE_STRINGS
                                | SHOW_ERROR_DETAIL
                                | MINIMAL_PARENTHESES
                                | MULTI_LINE_ADS
                                | MULTI_LINE_LISTS;

    /** Set options for converting expressions to strings.
     * @param flag the new flag values.
     * @see #NO_ESCAPE_STRINGS
     * @see #SHOW_ERROR_DETAIL
     * @see #MINIMAL_PARENTHESES
     * @see #MULTI_LINE_ADS
     * @see #MULTI_LINE_LISTS
     * @see #BRIEF
     * @see #COMPACT
     * @see #READABLE
     */
    static public void setFormatFlags(int flags) {
        formatFlags = flags;
    } // setFormatFlags(int)

    /** Get existing options for converting expressions to strings.
     * @return the current option flags.
     * @see #NO_ESCAPE_STRINGS
     * @see #SHOW_ERROR_DETAIL
     * @see #MINIMAL_PARENTHESES
     * @see #MULTI_LINE_ADS
     * @see #MULTI_LINE_LISTS
     * @see #BRIEF
     * @see #COMPACT
     * @see #READABLE
     */
    static public int getFormatFlags() {
        return formatFlags;
    } // getFormatFlags()

    /** Turn on options for converting expressions to strings.
     * @param flag the flag values to be enabled
     * @see #NO_ESCAPE_STRINGS
     * @see #SHOW_ERROR_DETAIL
     * @see #MINIMAL_PARENTHESES
     * @see #MULTI_LINE_ADS
     * @see #MULTI_LINE_LISTS
     * @see #BRIEF
     * @see #COMPACT
     * @see #READABLE
     */
    static public void enableFormatFlags(int flags) {
        formatFlags |= flags;
    } // enableFormatFlags(int)

    /** Turn off options for converting expressions to strings.
     * @param flag the flag values to be disabled
     * @see #NO_ESCAPE_STRINGS
     * @see #SHOW_ERROR_DETAIL
     * @see #MINIMAL_PARENTHESES
     * @see #MULTI_LINE_ADS
     * @see #MULTI_LINE_LISTS
     * @see #BRIEF
     * @see #COMPACT
     * @see #READABLE
     */
    static public void disableFormatFlags(int flags) {
        formatFlags &= ~flags;
    } // disableFormatFlags(int)

    /** Convert this Expr to a string.
     * The exact format of the result is controled by the static fields
     * <code>Expr.formatFlags</code> and <code>Expr.formatIndent</code>.
     * The result may be multiple lines long (it may include embedded newlines
     * and tabs), but it never ends with a newline.
     * @return a string representation of this Expr
     * @see #formatFlags
     * @see #formatIndent
     * @see #NO_ESCAPE_STRINGS
     * @see #SHOW_ERROR_DETAIL
     * @see #MINIMAL_PARENTHESES
     * @see #MULTI_LINE_ADS
     * @see #MULTI_LINE_LISTS
     * @see #BRIEF
     * @see #COMPACT
     * @see #READABLE
     */
    public abstract String toString();

    /** Convert this Expr to a string.
     * This is a convenience function that temporarily sets the static flags
     * <code>Expr.formatFlags</code> for the duration of a single call of the
     * no-argument version.
     * @param flags the new flags to set.
     * @return a string representation of this Expr
     * @see #toString
     * @see #formatFlags
     * @see #formatIndent
     * @see #NO_ESCAPE_STRINGS
     * @see #SHOW_ERROR_DETAIL
     * @see #MINIMAL_PARENTHESES
     * @see #MULTI_LINE_ADS
     * @see #MULTI_LINE_LISTS
     * @see #BRIEF
     * @see #COMPACT
     */
    public String toString(int flags) {
        int oldFlags = formatFlags;
        formatFlags = flags;
        String result = toString();
        formatFlags = oldFlags;
        return result;
    } // toString(int)

    /** The precedence of the operator in this expression node.
     * Used to print expressions without superfluous parentheses.
     * @return the precendence of this node.
     * @see #MINIMAL_PARENTHESES
     */
    abstract public int prec();

    /** The maximum precedence in the grammar for ClassAd expressions. */
    static protected final int MAXPREC = 11;

    /** Flag to control level of verbosity of debugging input.  Higher values
     * are more verbose.  If the value is zero, no debugging output is printed.
     */
    public static int dblevel = 0;

    /** Debugging print.
     * The message is printed to System.out.
     * Intended usage is something like
     * <pre>
     *    if (dblevel &gt; 1) db("The value of foo is " +foo);
     * </pre>
     * @param msg the message to be printed.
     * @see #dblevel
     */
    public static void db(Object msg) {
        System.out.println(msg);
    } // db(Object)

    /** Debugging print.
     * The message is printed to System.out.
     * @param newline if true, terminate the message with a newline.
     * @param msg the message to be printed.
     * @see #db(java.lang.Object)
     */
    public static void db(boolean newline, Object msg) {
        if (newline)
            System.out.println(msg);
        else
            System.out.print(msg);
    } // db(boolean,Object)

    /** A table mapping operator codes to character-string names.  For example,
     * <code>opName[NOT_EQUAL] = "!="</code>.
     */
    public static final String[] opName = {
        "||", "&&", "|", "^", "&", "==", "!=",
        " is ", " isnt ", "<", ">", "<=", ">=",
        "<<", ">>", ">>>", "+", "-", "*", "/",
        "%", "+", "-", "~", "!"
    }; // opName

    /** A table mapping operator codes to character-string names, escaped for
     * XML. For example,
     * <code>opName[LEFT_SHIFT] = "&amp;lt;&amp;lt;"</code>.
     */
    public static final String[] opNameXML = {
        "||", "&amp;&amp;", "|", "^", "&amp;", "==", "!=",
        " is ", " isnt ", "&lt;", "&gt;", "&lt;=", "&gt;=",
        "&lt;&lt;", "&gt;&gt;", "&gt;&gt;&gt;", "+", "-", "*", "/",
        "%", "+", "-", "~", "!"
    }; //opNameXML

    /** Token ID corresponding to the operator <b>||</b> */
    public static final int OR = 0;
    /** Token ID corresponding to the operator <b>&amp;&amp;</b> */
    public static final int AND = 1;
    /** Token ID corresponding to the operator <b>|</b> */
    public static final int BITOR = 2;
    /** Token ID corresponding to the operator <b>^</b> */
    public static final int BITXOR = 3;
    /** Token ID corresponding to the operator <b>&amp;</b> */
    public static final int BITAND = 4;
    /** Token ID corresponding to the operator <b>==</b> */
    public static final int EQUAL = 5;
    /** Token ID corresponding to the operator <b>!=</b> */
    public static final int NOT_EQUAL = 6;
    /** Token ID corresponding to the operator <b> is </b> */
    public static final int SAME = 7;
    /** Token ID corresponding to the operator <b> isnt </b> */
    public static final int DIFFERENT = 8;
    /** Token ID corresponding to the operator <b>&lt;</b> */
    public static final int LESS = 9;
    /** Token ID corresponding to the operator <b>&gt;</b> */
    public static final int GREATER = 10;
    /** Token ID corresponding to the operator <b>&lt;=</b> */
    public static final int LESS_EQ = 11;
    /** Token ID corresponding to the operator <b>&gt;=</b> */
    public static final int GREATER_EQ = 12;
    /** Token ID corresponding to the operator <b>&lt;&lt;</b> */
    public static final int LEFT_SHIFT = 13;
    /** Token ID corresponding to the operator <b>&gt;&gt;</b> */
    public static final int RIGHT_SHIFT = 14;
    /** Token ID corresponding to the operator <b>&gt;&gt;&gt;</b> */
    public static final int URIGHT_SHIFT = 15;
    /** Token ID corresponding to the operator <b>+</b> */
    public static final int PLUS = 16;
    /** Token ID corresponding to the operator <b>-</b> */
    public static final int MINUS = 17;
    /** Token ID corresponding to the operator <b>*</b> */
    public static final int TIMES = 18;
    /** Token ID corresponding to the operator <b>/</b> */
    public static final int DIV = 19;
    /** Token ID corresponding to the operator <b>%</b> */
    public static final int MOD = 20;
    /** Token ID corresponding to the operator <b>unary +</b> */
    public static final int UPLUS = 21;
    /** Token ID corresponding to the operator <b>unary -</b> */
    public static final int UMINUS = 22;
    /** Token ID corresponding to the operator <b>unary ~</b> */
    public static final int BIT_COMPLEMENT = 23;
    /** Token ID corresponding to the operator <b>unary !</b> */
    public static final int NOT = 24;
} // Expr
