/** $Header: /p/condor/repository/CONDOR_SRC/src_java/condor/classad/ListExpr.java,v 1.10 2002/07/20 21:21:29 roy Exp $ */

/*--------------------------------------------------------------------
 *
 * Condor Java ClassAd library
 * Copyright (C) 1990-2001, CONDOR Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and Marvin Solomon
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General
 * Public License as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *--------------------------------------------------------------------*/

package condor.classad;

import java.util.*;

/**
 * A list of expressions.
 * An internal (non-leaf) node of an expression tree representing the
 * list-construction operator <code>{ expr, expr, ... }</code>.
 * @see Expr
 * @author <a href="mailto:solomon@cs.wisc.edu">Marvin Solomon</a>
 * @version 1.2
 */
public class ListExpr extends Expr {
    private static String VERSION = "$Id: ListExpr.java,v 1.10 2002/07/20 21:21:29 roy Exp $";

    /** The component expressions of the list */
    private final ArrayList contents;

    /** Get the length of this list.
     * @return the number of elements in this list
     */
    public int size() {
        return contents.size();
    } // size()

    /** Construct an empty list. */
    public ListExpr() {
        super(LIST);
        contents = new ArrayList();
    } // ListExpr()

    /** Construct a list from a list of expressions.
     * @param contents the array of expressions.
     */
    public ListExpr(List l) {
        super(LIST);
        contents = new ArrayList(l);
    } // ListExpr(List)

    /** Get a member of this list.
     * @param i the member to be retrieved.
     * @return the ith member of this list, or an error Constant if i is out of
     * bounds.
     */
    public Expr sub(int i) {
        try {
            return (Expr) contents.get(i);
        } catch (IndexOutOfBoundsException e) {
            return Constant.error("subscript " + i + " is out of bounds");
        }
    } // sub(int)

    /** Appends the specified Expr to the end of this list.
     * @param e the Expr to add.
     * @return this ListExpr.
     */
    public Expr add(Expr e) {
        contents.add(e);
        return this;
    } // add(Expr)

    /** Get an iterator for iterating throught the members of this list.
     * @return an iterator for enumerating the Expr members of this list.
     */
    public Iterator iterator() {
        return contents.iterator();
    } // iterator

    /** Get the type of the expression.
     * @return the string "list".
     */
    protected String typeName() {
        return "list";
    } // typeName()

    /** Convert this Expr to a string.
     * The exact format of the result is controled by the static fields
     * <code>Expr.formatFlags</code> and <code>Expr.formatIndent</code>.
     * The result may be multiple lines long (it may include embedded newlines
     * and tabs), but it never ends with a newline.
     * @return a string representation of the this Context
     * @see Expr#formatFlags
     * @see Expr#formatIndent
     */
    public String toString() {
        String saveIndent, pre, sep, post;
        boolean multi = (formatFlags & MULTI_LINE_LISTS) != 0;

        if (multi) {
            saveIndent = formatIndent;
            formatIndent += "    ";
        } else {
            saveIndent = null;
        }

        if ((formatFlags & XML) != 0) {
            sep = multi ? ("\n" + formatIndent) : " ";
            pre = "<l>" + sep;
            post = sep + "</l>";
        } else {
            sep = multi ? (",\n" + formatIndent) : ", ";
            pre = multi ? ("{\n" + formatIndent) : "{ ";
            post = multi ? ("\n" + saveIndent + "}") : " }";
        }

        StringBuffer result = new StringBuffer(pre);
        for (int i = 0; i < contents.size(); i++) {
            if (i > 0) {
                result.append(sep);
            }
            result.append(contents.get(i).toString());
        }
        result.append(post);
        if (saveIndent != null) {
            formatIndent = saveIndent;
        }
        return result.toString();
    } // toString()

    /** The precedence of the operator in this expression node.
     * Used to print expressions without superfluous parentheses.
     * @return the precendence of this node.
     * @see Expr#MINIMAL_PARENTHESES
     */
    public int prec() {
        return MAXPREC+1;
    } // prec()
} // ListExpr
