/** $Header: /p/condor/repository/CONDOR_SRC/src_java/condor/classad/Query.java,v 1.6 2002/07/20 21:21:29 roy Exp $ */

/*--------------------------------------------------------------------
 *
 * Condor Java ClassAd library
 * Copyright (C) 1990-2001, CONDOR Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and Marvin Solomon
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General
 * Public License as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *--------------------------------------------------------------------*/

package condor.classad;

import java.util.*;
import java.io.*;

/** A parsed "query" in a very simple ad hoc query language inspired by SQL.
 * This is a hack desiged to test the ClassAd machinery by sending a query
 * to Condor Collector and interpreting the results.
 * A query has the form
 * <pre>
 *     Query: 'select' SelectClause 'from' Identifier WhereClause
 *     SelectClause: '*' | SelectionList
 *     WhereClause : [ 'where' Expression ]
 *     SelectionList: [ SelectionList ',' ] Identifier [ '=' Expression ]
 * </pre>
 */
public class Query {
    private static String VERSION = "$Id: Query.java,v 1.6 2002/07/20 21:21:29 roy Exp $";

    /** The "select" clause, encoded as a RecordExpr.
     * The attribute names are the supplied &lt;IDENTIFIERS&gt;, or generated
     * names of the form ``$argn'' if the &lt;IDENTIFIER&gt; is omitted.
     */
    public RecordExpr selectClause;

    /** The "from" clause (as a string). */
    public String fromClause;

    /** The "where" clause (as an expression of the form [Requirements = ...]),
     * or null if the "where" clause was omitted.
     */
    public RecordExpr whereClause;

    /** Construct a query from its components.
     * @param s the "select" clause.
     * @param f the "from" clause.
     * @param w the "where" clause (or null).
     */
    public Query(RecordExpr s, String f, RecordExpr w) {
        selectClause = s;
        fromClause = f;
        whereClause = w;
    }

    /** Create a Query by parsing an input stream.
     * @param s the stream to parse.
     * @return a new Query object.
     */
    public static Query parse(InputStream s) {
        ClassAdParser parser = new ClassAdParser(s);
        return parser.parseQuery();
    }

    /** Create a Query my parsing an input stream.
     * @param s the stream to parse.
     * @return a new Query object.
     */
    public static Query parse(Reader s) {
        ClassAdParser parser = new ClassAdParser(s);
        return parser.parseQuery();
    }
    /** Create a Query my parsing an input stream.
     * @param s the string to parse.
     * @return a new Query object.
     */
    public static Query parse(String s) {
        ClassAdParser parser = new ClassAdParser(s);
        return parser.parseQuery();
    }
    public String toString() {
        String result;
        if (selectClause == null)
            result = "select *";
        else {
            result = "select";
            String sep = " ";
            for (Iterator e = selectClause.attributes(); e.hasNext(); ) {
                result += sep;
                sep = ", ";
                String attr = (String)e.next();
                if (attr.charAt(0) != '$')
                    result += attr + " = ";
                result += selectClause.lookup(attr).toString(Expr.READABLE);
            }
        }
        result += "\nfrom " + fromClause;
        if (whereClause != null)
            result += "\nwhere "
                + whereClause.lookup("Requirements")
                    .toString(Expr.MINIMAL_PARENTHESES);
        return result;
    }
} // Query
