/** $Header: /p/condor/repository/CONDOR_SRC/src_java/condor/classad/RecordExpr.java,v 1.17 2002/07/20 21:21:29 roy Exp $ */

/*--------------------------------------------------------------------
 *
 * Condor Java ClassAd library
 * Copyright (C) 1990-2001, CONDOR Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and Marvin Solomon
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General
 * Public License as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *--------------------------------------------------------------------*/

package condor.classad;
import java.util.*;
import java.io.*;

/**
 * A "record" expression (also known as a "classad").
 * An internal (non-leaf) node of an expression tree representing a classified
 * ad (a.k.a. property list, record, table, map, etc.)
 * @see Expr
 * @author <a href="mailto:solomon@cs.wisc.edu">Marvin Solomon</a>
 * @version 1.2
 */
public class RecordExpr extends Expr {
    private static String VERSION = "$Id: RecordExpr.java,v 1.17 2002/07/20 21:21:29 roy Exp $";

    /** The immediately enclosing RecordExpr (null if this is the top level).
     */
    private RecordExpr parent;

    /** A counter for generating unique attribute names. */
    private int nextAttribute = 0;

    /** Default constructor: a record with no fields */
    public RecordExpr() {
        super(RECORD);
    } // RecordExpr()

    /** A mapping from CiString attribute names to Expr values.  */
    private Map map = new TreeMap();

    /** A list of all the attribute names.  These strings can also be obtained
     * by enumerating the keys of <code>map</code>, but for debugging purposes,
     * we would like to maintain them in the order they were added.
     * Elements of this list are intern'ed strings in the original case they
     * were supplied.
     */
    private List attrNames = new LinkedList();

    /** The special (reserved) ``super'' attribute.  */
    private CiString superKey = CiString.getInstance("super");

    /** Serialize this RecordExpr in the format currently expected by Condor.
     * <ul>
     *    <li> an integer n
     *    <li> n+2 strings representing all attributes.  All but the last two
     *         have the form "name = value", where value is the value of the
     *         attribute converted to a string by toString (currently,
     *         toString(COMPACT)).  The last two are the toString(BRIEF)
     *         verisons of attributes MyType and TargetType, in that order,
     *         which must be present in the ad.
     * </ul>
     * The integer n is serialized by the writeInt method of the output stream
     * and the n+2 strings are serialized by the writeUTF method.  Normally,
     * The output stream will be an instance of condor.cedar.CedarOutputStream,
     * which means that n is written as 8 bytes, most significant byte first,
     * and the strings are null-terminated and encoded using Java UTF-8 (which
     * just transmits the low 8 bits of each character provided the string
     * contains only ASCII characters).
     * <p>
     * To reconstitute a classad from this format, use a {@link ClassAdReader}
     * and a {@link ClassAdParser}.
     * @param out the stream to which the serialzed data should be written.
     * @return the byte-array result.
     * @exception InvalidObjectException if this add does not have MyType and
     * TargetType attributes.
     * @exception IOException if an I/O error occurs.
     * @see ClassAdParser#parse()
     * @see ClassAdReader
     */
    public void transmit(DataOutput out) throws IOException {
        Expr myType = lookup("MyType");
        if (myType == null) {
            throw new InvalidObjectException("no MyType attribute");
        }
        Expr targetType = lookup("TargetType");
        if (targetType == null) {
            throw new InvalidObjectException("no TargetType attribute");
        }
        out.writeInt(map.size() - 2);
        for (Iterator keys = attrNames.iterator();keys.hasNext();) {
            String name = (String)keys.next();
            if (name.equals("MyType") || name.equals("TargetType"))
                continue;
            Expr value = lookup(name);
            out.writeUTF(name + "=" + value.toString(COMPACT));
        }
        out.writeUTF(myType.toString(BRIEF));
        out.writeUTF(targetType.toString(BRIEF));
    } // transmit(DataOutput)

    /** The type of the expression.
     * @return the string ``classad''.
     */
    protected String typeName() {
        return "classad";
    } // typeName()

    /** Find the attribute with the given name.
     * @param name the attribute name to look for.
     * @return the value of the attribute (null if not present).
     */
    public Expr lookup(String name) {
        return (Expr)map.get(CiString.getInstance(name));
    } // lookup(String)

    /** Find the attribute with the given name.
     * @param name the attribute name to look for.
     * @return the value of the attribute (null if not present).
     */
    public Expr lookup(CiString name) {
        return (Expr)map.get(name);
    } // lookup(CiString)

    /** Find an attribute with the given name in this RecordExpr or a
     * ``containing'' one.  The ``containing'' ad may be the staticly enclosing
     * ad or one found in the context.
     * @param ctx a context for evaluating any ``super'' attribute encountered.
     * @param name the attribute name to look for.
     * @return the value of the attribute (null if not present).
     */
    public Expr search(String name, Context ctx) {
        return search(CiString.getInstance(name), ctx);
    } // search(String,Context)

    /** Find an attribute with the given name in this RecordExpr or a
     * ``containing'' one.  The ``containing'' ad may be the staticly enclosing
     * ad or one found in the context.
     * @param ctx a context for evaluating any ``super'' attribute encountered.
     * @param key the attribute name to look for.
     * @return the value of the attribute (null if not present).
     */
    public Expr search(CiString key, Context ctx) {
        Expr result = null;
        RecordExpr ad = this;
        Set seen = new HashSet();

        while (ad != null) {
            if (dblevel>=3) db(formatIndent + "Search for key " + key
                + " in ad " + ad);
            result = (Expr)ad.lookup(key);
            if (result != null || seen.contains(ad))
                break;
            seen.add(ad);
            if (dblevel>=3) db(formatIndent + "parent=" + parent
                + "; ctx.getParent=" + ctx.getParent(ad));
            ad = ad.parent != null ? parent : ctx.getParent(ad);
        }
        if (dblevel>=3) db(formatIndent + "Search returns " + result);
        return result;
    } // search(CiString,Context)

    /** Add an attribute (replacing previous value if any)
     * @param name the name of the attribute to be added.
     * @param expr the value of the attribute.
     * @return this RecordExpr.
     * @throws IllegalArgumentException if name or expr is null.
     */
    public RecordExpr insertAttribute(String name, Expr expr) {
        return insertAttribute(CiString.getInstance(name), expr);
    } // insertAttribute(String,Expr)

    /** Add an attribute (replacing previous value if any)
     * @param key the name of the attribute to be added.
     * @param expr the value of the attribute.
     * @return this RecordExpr.
     * @throws IllegalArgumentException if name or expr is null.
     */
    public RecordExpr insertAttribute(CiString key, Expr expr) {
        if (key == null) {
            throw new IllegalArgumentException(
                        "null attribute name " + key);
        }
        if (expr == null) {
            throw new IllegalArgumentException(
                        "null value for attribute " + key);
        }
        if (!map.containsKey(key))
            attrNames.add(key.toString());
        map.put(key, expr);
        return this;
    } // insertAttribute(CiString,Expr)

    /** Add an attribute with a generated name.  The name has the form
     * $arg&lt;<em>nn</em>&gt;, where <em>nn</em> is a non-negative integer.
     * @param expr the value of the attribute.
     * @return this RecordExpr.
     * @throws IllegalArgumentException if expr is null.
     */
    public RecordExpr insertAttribute(Expr expr) {
        String name = "$arg" + nextAttribute;
        nextAttribute++;
        return insertAttribute(CiString.getInstance(name), expr);
    } // insertAttribute(Expr)

    /** Count how many attributes.
     * @return the number of distinct attributes in this ClassAd.
     */
    public int size() {
        return map.size();
    } // size()

    /** Enumerate the attribute names.
     * @return an iterator of objects of type String, representing the
     *    the attribute names.
     */
    public Iterator attributes() {
        return attrNames.iterator();
    } // attributes()

    /** Enumerate the attribute values.
     * @return an iterator of objects of type Expr, representing the
     *    the attribute values.
     */
    public Iterator values() {
        return map.values().iterator();
    } // attributes()

    /** Convert this Expr to a string.
     * The exact format of the result is controled by the static fields
     * <code>Expr.formatFlags</code> and <code>Expr.formatIndent</code>.
     * The result may be multiple lines long (it may include embedded newlines
     * and tabs), but it never ends with a newline.
     * @return a string representation of the this Context
     * @see Expr#formatFlags
     * @see Expr#formatIndent
     */
    public String toString() {
        String saveIndent, pre, sep, post, pre1, sep1, post1;
        boolean multi = (formatFlags & MULTI_LINE_ADS) != 0;

        if (multi) {
            saveIndent = formatIndent;
            formatIndent += "    ";
        } else {
            saveIndent = null;
        }

        if ((formatFlags & XML) != 0) {
            sep = multi ? ("\n" + formatIndent) : " ";
            pre = "<ca>" + sep;
            post = sep + "</ca>";
            pre1 = "<a n=\"";
            sep1 = "\">";
            post1 = "</a>";
        } else {
            sep = multi ? (";\n" + formatIndent) : "; ";
            pre = multi ? ("[\n" + formatIndent) : "[ ";
            post = multi ? ("\n" + saveIndent + "]") : " ]";
            pre1 = "";
            sep1 = " = ";
            post1 = "";
        }

        StringBuffer result = new StringBuffer(pre);
        boolean firstTime = true;
        for (Iterator keys = attrNames.iterator(); keys.hasNext();) {
            String name = (String)keys.next();
            if (firstTime) {
                firstTime = false;
            } else {
                result.append(sep);
            }
            result.append(pre1)
                .append(name)
                .append(sep1)
                .append(lookup(name).toString())
                .append(post1);
        }
        result.append(post);
        if (saveIndent != null) {
            formatIndent = saveIndent;
        }
        return result.toString();
    } // toString()

    /** The precedence of the operator in this expression node.
     * Used to print expressions without superfluous parentheses.
     * @return the precendence of this node.
     * @see Expr#MINIMAL_PARENTHESES
     */
    public int prec() {
        return MAXPREC;
    } // prec()
} // RecordExpr
