/** $Header: /p/condor/repository/CONDOR_SRC/src_java/condor/classad/SelectExpr.java,v 1.12 2002/07/20 21:21:29 roy Exp $ */

/*--------------------------------------------------------------------
 *
 * Condor Java ClassAd library
 * Copyright (C) 1990-2001, CONDOR Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and Marvin Solomon
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General
 * Public License as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *--------------------------------------------------------------------*/

package condor.classad;

import java.util.*;

/**
 * An internal (non-leaf) node of an expression tree, representing a selection
 * of component of a record.  It has the form from a class ad in the form
 * <pre>
 *    &lt;expr&gt; . &lt;identifier&gt;.
 * </pre>
 * @see Expr
 * @author <a href="mailto:solomon@cs.wisc.edu">Marvin Solomon</a>
 * @version 1.2
 */
public class SelectExpr extends Expr {
    private static String VERSION = "$Id: SelectExpr.java,v 1.12 2002/07/20 21:21:29 roy Exp $";

    /** The right argument (identifier) of the selection operator. */
    public final CiString selector;

    /** The left argument (base expression) of the selection operator. */
    public final Expr base;

    /** Create a new SelectExpr corresponding to <code>base.selector</code>.
     * @param base the left argument (base expression).
     * @param selector the right argument (identifier).
     */
    public SelectExpr(Expr base, String selector) {
        this(base, CiString.getInstance(selector));
    } // SelectExpr(Expr,String)

    /** Create a new SelectExpr corresponding to <code>base.selector</code>.
     * @param base the left argument (base expression).
     * @param selector the right argument (identifier).
     */
    public SelectExpr(Expr base, CiString selector) {
        super(SELECTION);
        this.base = base;
        this.selector = selector;
    } // SelectExpr(Expr,CiString)

    /** A convenience function for creating a SelectExpression corresponding
     * to "base.a.b.c"
     * @param base the base Expr.
     * @param sel the sequence of selectors <code>{"a", "b", "c"}</code>
     * @return a new SelectExpr or <code>base</code> if <code>sel</code> is
     * empty.
     */
    public static Expr select(Expr base, String[] sel) {
        for (int i=0; i<sel.length; i++)
            base = new SelectExpr(base, CiString.getInstance(sel[i]));
        return base;
    } // select(Expr,String)

    /** The type of the expression.  This should never be called!
     * @return the string "selection".
     */
    protected String typeName() {
        return "record";
    } // typeName()

    /** Convert this Expr to a string.
     * The exact format of the result is controled by the static fields
     * <code>Expr.formatFlags</code> and <code>Expr.formatIndent</code>.
     * The result may be multiple lines long (it may include embedded newlines
     * and tabs), but it never ends with a newline.
     * @return a string representation of the this Context
     * @see Expr#formatFlags
     * @see Expr#formatIndent
     */
    public String toString() {
        StringBuffer result = new StringBuffer();
        int flags = formatFlags;
        if ((flags & XML) != 0) {
            formatFlags = formatFlags & ~XML | XML_CONTENT;
            result.append("<e>");
        }
        if (base.type == ATTRIBUTE) {
            // Special kludge to replace "this.sel" with the more readable "sel"
            // and "root.sel" with ".sel"
            switch (((AttrRef)base).locality) {
                case AttrRef.SELF:
                    result.append(mark()).append(selector);
                    break;
                case AttrRef.ROOT:
                    result.append(mark()).append(".").append(selector);
                    break;
                default: break;
            }
        } else {
            result.append(base).append(mark()).append(".").append(selector);
        }
            
        if ((flags & XML) != 0) {
            formatFlags = flags;
            result.append("</e>");
        }
        return result.toString();
    } // toString()

    /** The precedence of the operator in this expression node.
     * Used to print expressions without superfluous parentheses.
     * @return the precendence of this node.
     * @see Expr#MINIMAL_PARENTHESES
     */
    public int prec() {
        return MAXPREC+1;
    } // prec()

    /** Evalutate this expression.
     * @param ctx an evaluation context for evaluating this expression
     * @return a Constant, ListExpr, or RecordExpr representing the value.
     * @see Expr#eval(condor.classad.Context)
     */
    protected Expr reEval(Context ctx) {
        Expr b = base.eval(ctx);
        Expr result;
        if (b.type == UNDEFINED || b.type == ERROR)
            result = b;
        else if (b.type == RECORD) {
            result = ((RecordExpr)b).search(selector, ctx);
            if (result == null)
                result = Constant.undefined("attribute " + selector
                                        + " not found in expression " + b);
        }
	else if (b.type == LIST) {
	    ListExpr evaledList = new ListExpr();
	    for(Iterator bItr = ((ListExpr)b).iterator(); bItr.hasNext();){
		Expr currentExpr = (Expr)bItr.next();
		SelectExpr selExpr = new SelectExpr( currentExpr, selector );
		evaledList.add(selExpr.eval(ctx));
	    }
	    result = evaledList;
	}
        else {
            result = Constant.error(
                        b + "." + selector + ": argument is not a record");
        }
        return result;
    } // reEval(Context)
} // SelectExpr
