/** $Header: /p/condor/repository/CONDOR_SRC/src_java/condor/classad/SubscriptExpr.java,v 1.9 2002/07/20 21:21:29 roy Exp $ */

/*--------------------------------------------------------------------
 *
 * Condor Java ClassAd library
 * Copyright (C) 1990-2001, CONDOR Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and Marvin Solomon
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General
 * Public License as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *--------------------------------------------------------------------*/

package condor.classad;

/**
 * A selection from a list.
 * An internal (non-leaf) node of an expression tree representing a subscript
 * applied to another expression (which presumedly evaluates to a list).
 * @see Expr
 * @author <a href="mailto:solomon@cs.wisc.edu">Marvin Solomon</a>
 * @version 1.2
 */
public class SubscriptExpr extends Expr {
    private static String VERSION = "$Id: SubscriptExpr.java,v 1.9 2002/07/20 21:21:29 roy Exp $";

    /** The left argument (base expression) of the subscript operator. */
    public final Expr base;

    /** The right argument (subscript) of the selection operator. */
    public final Expr selector;

    /** Create a new SubscriptExpr.
     * @param base the left argument (base expression).
     * @param selector the right argument (subscript).
     */
    public SubscriptExpr(Expr base, Expr selector) {
        super(SUBSCRIPT);
        this.base = base;
        this.selector = selector;
    } // SubscriptExpr(Expr,Expr)

    /** The type of the expression.  This should never be called!
     * @return the string "selection".
     */
    protected String typeName() {
        return "subscript";
    } // typeName()

    /** Convert this Expr to a string.
     * The exact format of the result is controled by the static fields
     * <code>Expr.formatFlags</code> and <code>Expr.formatIndent</code>.
     * The result may be multiple lines long (it may include embedded newlines
     * and tabs), but it never ends with a newline.
     * @return a string representation of the this Context
     * @see Expr#formatFlags
     * @see Expr#formatIndent
     */
    public String toString() {
        if ((formatFlags & XML) != 0) {
            int save = formatFlags;
            formatFlags = formatFlags & ~XML | XML_CONTENT;
            String b = base.toString();
            String s = selector.toString();
            String result = "<e>" + b + "[" + s + "]</e>";
            formatFlags = save;
            return result;
        } else {
            return base + "[" + selector + "]";
        }
    } // toString()

    /** The precedence of the operator in this expression node.
     * Used to print expressions without superfluous parentheses.
     * @return the precendence of this node.
     * @see Expr#MINIMAL_PARENTHESES
     */
    public int prec() {
        return MAXPREC;
    } // prec()

    /** Evalutate this expression.
     * @param ctx an evaluation context for evaluating this expression
     * @return a Constant, ListExpr, or RecordExpr representing the value.
     * @see Expr#eval(condor.classad.Context)
     */
    protected Expr reEval(Context ctx) {
        Expr b = base.eval(ctx);
        Expr i = selector.eval(ctx);
        if (b.type != LIST)
            return Constant.error("[]: object is " + b.typeName()
                + ", not a list");
        if (i.type != INTEGER)
            return Constant.error("[]: subscript is " + i.typeName()
                + ", not an integer");
        return ((ListExpr)b).sub(((Constant)i).intValue());
    } // reEval(Context)
} // SubscriptExpr
