/** $Header: /p/condor/repository/CONDOR_SRC/src_java/condor/classad/tests/ExprTest.java,v 1.2 2002/07/20 21:21:36 roy Exp $ */

/*--------------------------------------------------------------------
 *
 * Condor Java ClassAd library
 * Copyright (C) 1990-2001, CONDOR Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and Marvin Solomon
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General
 * Public License as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *--------------------------------------------------------------------*/

package condor.classad.tests;
import condor.classad.*;
import java.io.*;
import java.util.*;

/** Main program to test evaluation of expressions.
 * Input consists of a sequence of expressions, each followed by a semicolon.
 * Each expression is one of the following:
 * <ul>
 * <li>A list of two classads (RECORD expressions); in this case ad1.value is
 * evaluated in the same environment used for matching classads.
 * <li>A single classad (RECORD) expression; in this case, ad.value is
 * evaluated.
 * <li>Anything else; the program terminates.
 * </ul>
 */
public class ExprTest {
    /** Abbreviation for system.out.println.
     * @param s something to be printed (generally a String)
     */
    private static void pl(Object s) {
        System.out.println(s);
    }

    /** Main program.
     * Command-line options are
     * <dl>
     * <dt>-v<dd>Request more verbose debugging output (may be repeated).
     * <dt>-v<em>nn</em><dd>Set the debugging level to <em>nn</em>.
     * <dt>-t<dd>Trace the actions of the parser.
     * </dl>
     * An optional file name may be supplied (default is stdin).
     */
    public static void main(String[] args) throws Exception {
        new ExprTest().run(args);
    }

    /** Print a usage message and exit. */
    private static void usage() {
        System.err.println(
            "usage: java ExprTest [-v[verbosity]] [-t] [fname]");
        System.exit(1);
    }
    public void run(String[] args) throws Exception {
        // Get command-line arguments
        int verbosity = 0; // amount of debugging output
        boolean tracing = false; // -t flag:  tell the parser to trace
        GetOpt opts = new GetOpt(getClass().getName(), args, "v::t");
    opt_loop:
        for (;;) switch(opts.nextOpt()) {
            case 'v': // More verbose
                if (opts.optarg == null) verbosity++;
                else verbosity = Integer.parseInt(opts.optarg);
                break;
            case 't': // parser trace
                tracing = true;
                break;
            case -1:
                break opt_loop;
            default: usage();
        }

        Reader input = null;
        if (opts.optind < args.length) {
            try {
                input = new FileReader(args[opts.optind]);
            } catch (Exception e) {
                System.err.println(args[opts.optind] + ": " + e);
                usage();
            }
        }
        else
            input = new InputStreamReader(System.in);

        Expr.setFormatFlags(
            Expr.MULTI_LINE_ADS
            | Expr.MULTI_LINE_LISTS
            | Expr.MINIMAL_PARENTHESES );
        ClassAdParser parser = new ClassAdParser(input);
        parser.enableTracing(tracing);

        // main loop
        for(;;) {
            Expr expr = parser.parse();
            if (expr==null)
                break;
            if (expr.type == Expr.LIST) {
                ListExpr l = (ListExpr)expr;
                if (l.size() != 2) {
                    System.out.println("Bad arg '" + expr + "':");
                    System.out.println("List with " + l.size()
                        + " elements (should be 2)");
                    continue;
                }
                Expr e1 = l.sub(0);
                Expr e2 = l.sub(1);
                System.out.println("A = " + e1);
                System.out.println("B = " + e2);
                if (e1.type != Expr.RECORD || e2.type != Expr.RECORD) {
                    System.out.println("Bad arg '" + expr + "':");
                    System.out.println("Non-ClassAd in list");
                    continue;
                }
                Expr val = ClassAd.eval("value",
                                (RecordExpr)e1, (RecordExpr)e2);
                System.out.println("result = " + val.toString(Expr.READABLE));
                System.out.println();
            }
            else if (expr.type == Expr.RECORD) {
                System.out.println("A = " + expr);
                Expr val = ClassAd.eval((RecordExpr)expr,
                                new String[] {"value"});
                System.out.println("result = " + val.toString(Expr.READABLE));
                System.out.println();
            }
            else return;
            int lookahead = parser.getNextToken();
            if (lookahead != ';') {
                System.out.println("Expecting semicolon, found " + lookahead);
                return;
            }
        } // for (;;)
    } // run
} // class ExprTest
