// $Header: /p/condor/repository/CONDOR_SRC/src_java/condor/classad/tests/QueryTest.java,v 1.10 2002/07/20 21:21:36 roy Exp $

/*--------------------------------------------------------------------
 *
 * Condor Java ClassAd library
 * Copyright (C) 1990-2001, CONDOR Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and Marvin Solomon
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General
 * Public License as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *--------------------------------------------------------------------*/

package condor.classad.tests;

import condor.classad.*;
import condor.cedar.*;
import java.io.*;
import java.util.*;
import java.net.*;
import java.util.Date;

/** Main program to exercise the matching feature of classads.
 * It reads ads from a file in the same format as the output of Fetch, runs a
 * query in an SQL-like langauge on them, and prints the results.  The query
 * comes either from the command line or from standard input.  It looks like
 * this:
 * <pre>
 *    select Name, Arch, Memory, Disk, Idle = timeInterval(ConsoleIdle)
 *    from ads_bin
 *    where ConsoleIdle &gt; 24*60*60
 * </pre>
 * This class is meant to be an example of programming with classads, to be
 * studied and adapted.  It should not be used as a black box.
 * @author <a href="mailto:solomon@cs.wisc.edu">Marvin Solomon</a>
 */
public class QueryTest {

    /** Collector port. */
    public static final int COLLECTOR_COMM_PORT = 9618;

    /** Abbreviation for system.out.println.
     * @param s something to be printed (generally a String)
     */
    private static void pl(Object s) {
        System.out.println(s);
    }

    /** Abbreviation for system.out.print.
     * @param s something to be printed (generally a String)
     */
    private static void p(Object s) {
        System.out.print(s);
    }

    /** Main program.
     * Command-line options are
     * <dl>
     * <dt>-v<dd>Request more verbose debugging output (may be repeated).
     * <dt>-v<em>nn</em><dd>Set the debugging level to <em>nn</em>.
     * <dt>-m<em>nn</em><dd>Limit the number of items printed by "select *"
     *           to <em>nn</em>.
     * <dt>-c<dd>Get the ads directly from the Condor collector rather than
     *           readting them from a file (and ignore the "from" clause of the
     *           query).
     * <dt>-s<dd>Let the server do the filtering:  Send the "where" clause as
     *           part of the query to the collector, so that only matching ads
     *           are returned.  (Implies -c).
     * </dl>
     * Any remaining words on the command line are assumed to be query.
     * If there are no non-flag arguments, the query is read from stdin.
     * @param args the command-line arguments.
     */
    public static void main(String args[]) {
        // ClassAdParser.disable_tracing();
        new QueryTest().run(args);
    }

    /** Print a usage message and exit. */
    private void usage() {
        System.err.println("usage: java "+getClass().getName()
            +" [-v [debug_level]] [-m count] [-c] [-s] [query]");
        System.exit(1);
    }

    /** The main body of the program.
     * @param args the command-line arguments.
     */
    private void run(String args[]) {
        // Get command-line arguments
        int maxads = -1; // number of ads to print with select * (-1 = unlimted)
        int verbosity = 0; // amount of debugging output
        boolean cflag = false; // get ads from collector rather than file
        boolean sflag = false; // do filtering at server (the collector)
        String server = "condor.cs.wisc.edu";
        GetOpt opts = new GetOpt(getClass().getName(), args, "v::m:cs");
    opt_loop:
        for (;;) switch(opts.nextOpt()) {
            case 'v': // More verbose
                if (opts.optarg == null) verbosity++;
                else verbosity = Integer.parseInt(opts.optarg);
                break;
            case 'm': // max ads
                maxads = Integer.parseInt(opts.optarg);
                break;
            case 'c': // use the collector
                cflag = true;
                break;
            case 's': // do the "where" filtering on the server
                sflag = true;
                cflag = true;
                break;
            case -1:
                break opt_loop;
            default: usage();
        }

        // Parse the query
        Query q = null;
        if (opts.optind >= args.length)
            q = Query.parse(System.in);
        else {
            String s = args[opts.optind];
            for (opts.optind++;opts.optind<args.length;opts.optind++)
                s += " " + args[opts.optind];
            q = Query.parse(s);
        }
        if (q == null) {
            pl("Cannot parse query");
            return;
        }

        // "Run" the query:  If the query is
        //    select .... from fname where expr
        // the ads are read from file "fname" (or directly from the Condor
        // collector) and put into a Vector if they satisfy the given expr.
        // All ads are kept if "where expr" is omitted.
        Vector results = new Vector();
        String adsFile = q.fromClause;
        Reader adsReader;
        if (verbosity > 0 && q.whereClause != null)
            Expr.db("Filter by " + q.whereClause.toString(Expr.READABLE));
        int totalAds = 0;
        try {
            if (cflag) {
                if (verbosity > 0)
                    pl("Connecting to host '" + server
                    + "', port " + COLLECTOR_COMM_PORT);
                Socket sock = new Socket(server, COLLECTOR_COMM_PORT);
                if (verbosity > 0)
                    pl("Connected: " + sock);
                CedarOutputStream out
                    = new CedarOutputStream(sock.getOutputStream());
                RecordExpr ad = new RecordExpr();
                ad.insertAttribute("Rank", ClassAd.constant(0));
                ad.insertAttribute("MyType", ClassAd.constant("Query"));
                ad.insertAttribute("TargetType", ClassAd.constant("machine"));
                if (sflag) {
                    RecordExpr w = q.whereClause;
                    ad.insertAttribute(
                        "Requirements",
                        w.lookup("Requirements"));
                }
                out.writeInt(5); // QUERY_STARTD_ADS
                ad.transmit(out);
                out.endOfMessage();
                adsReader
                    = new ClassAdReader(
                        new CedarInputStream(
                            new BufferedInputStream(
                                sock.getInputStream())));
            } else {
                adsReader
                    = new InputStreamReader(
                        new BufferedInputStream(
                            new FileInputStream(
                                adsFile)));
            }
            ClassAdParser parser = new ClassAdParser(adsReader);

            if (verbosity > 0)
                pl("Parsing ads");
            Date start = new Date();
            for (;;) {
                Expr ad = parser.parse();
                if (ad == null)
                    break;
                totalAds++;
                int lookahead = parser.getNextToken();
                if (lookahead != ';') {
                    pl("Expecting semicolon, found " + lookahead);
                    break;
                }
                if (q.whereClause != null) {
                    if (verbosity == 1)
                        p(".");
                    if (ClassAd.oneWayMatch(q.whereClause, ad)) {
                        if (verbosity > 1)
                            Expr.db("select " + ad.toString(Expr.READABLE));
                        results.addElement(ad);
                    }
                    else {
                        if (verbosity > 1)
                            Expr.db("reject " + ad.toString(Expr.READABLE));
                    }
                }
                else {
                    if (verbosity > 0)
                        p("*");
                    results.addElement(ad);
                }
            }
            long elapsed = (new Date()).getTime() - start.getTime();
            int bytesRead;
            if (cflag) {
                bytesRead = ((ClassAdReader)adsReader).getByteCount();
            } else {
                bytesRead = (int) (new File(adsFile)).length();
            }
            float rate = (float)((double)(1000*bytesRead)/elapsed);
            if (verbosity > 0) {
                pl("");
                pl(bytesRead + " bytes read in " + elapsed + " ms ("
                    + rate + " bps)");
            }
        }
        catch (FileNotFoundException e) {
            pl(e.getMessage() + ": no such file");
            return;
        }
        catch (Exception e) {
            e.printStackTrace(System.err);
            return;
        }
        if (verbosity > 0) {
            pl("Total ads received = " + totalAds);
        }

        // Now pretty-print the result.  If the select clause is "select *",
        // then simply dump the resulting ads.  Otherwise, arrange them in
        // nice rows and columns.

        if (q.selectClause == null) {
            // 'select * from ...'
            // Just dump all the results
            int count = 0;
            for (Iterator res = results.iterator(); res.hasNext();) {
                Expr ad = (RecordExpr)res.next();
                pl(ad.toString(Expr.READABLE));
                count++;
                if (maxads >= 0 && count >= maxads)
                    break;
            }
            return;
        }

        // We make two passes over the result tuples.
        // On the first pass, we determine the type of each selected field
        // (for now, fields are either integer or "other"),
        // find the max length of each field, and accumulate totals.
        // On the second pass, we print the results.

        if (verbosity > 0)
            pl("Analyzing " + results.size() + " rows...");

        // Initialize tables with stuff that is the same for all tuples in the
        // result.
        RecordExpr fields = q.selectClause;
        int     fieldCount = fields.size();
        String  fieldName[] = new String[fieldCount];
        Expr    fieldExpr[] = new Expr[fieldCount];
        int     maxLen[] = new int[fieldCount];
        int     type[] = new int[fieldCount];
                        // -1 means undefined
                        //  0 means type not known yet
                        //  1 means int
                        //  2 means other
        int     sum[] = new int[fieldCount];
        int     lineLen = 0;
        Iterator attrs = fields.attributes();
        for (int i=0; i<fieldCount; i++) {
            String attr = (String)attrs.next();
            fieldExpr[i] = fields.lookup(attr);
            if (attr.charAt(0) == '$')
                attr = fieldExpr[i].toString();
            fieldName[i] = attr;
            maxLen[i] = fieldName[i].length();
            type[i] = 0;
            sum[i] = 0;
        }

        // Pass 1: check types and lengths and accumulate totals
        for (Iterator res = results.iterator(); res.hasNext();) {
            RecordExpr ad = (RecordExpr)res.next();
            Context ctx = ClassAd.bind(fields, ad);
            for (int i=0; i<fieldCount; i++) {
                Expr value = fieldExpr[i].eval(ctx);
                int ty = 2;
                switch (value.type) {
                    case Expr.INTEGER:
                        sum[i] += ((Constant)value).intValue();
                        ty = 1;
                        break;
                    case Expr.UNDEFINED:
                        ty = -1;
                        break;
                }
                    
                if (ty >= 0) {
                    if (type[i] == 0) {
                        type[i] = ty;
                    }
                    else if (type[i] != ty) {
                        pl("Error: mixed types for attribute "+fieldName[i]);
                        return;
                    }
                }
                int len = value.toString(Expr.BRIEF).length();
                if (len > maxLen[i])
                    maxLen[i] = len;
            }
        } // End of Pass 1

        // Calculate line length
        for (int i=0; i<fieldCount; i++) {
            if (type[i]==1) {
                int len = (""+sum[i]).length();
                if (len > maxLen[i])
                    maxLen[i] = len;
            }
            lineLen += maxLen[i];
        }
        lineLen += 3*(fieldCount-1); // for field separators

        // Print headers
        for (int i=0; i<fieldCount; i++) {
            if (i>0)
                p(" | ");
            field(fieldName[i], maxLen[i], type[i]==1);
        }
        pl("");
        for (int i=0; i<lineLen; i++)
            p("-");
        pl("");

        // Pass 2: print results
        for (Iterator res = results.iterator(); res.hasNext();) {
            RecordExpr ad = (RecordExpr)res.next();
            Context ctx = ClassAd.bind(fields, ad);
            for (int i=0; i<fieldCount; i++) {
                if (i>0)
                    p(" | ");
                Expr value = fieldExpr[i].eval(ctx);
                String s = value.toString(Expr.BRIEF);
                field(s, maxLen[i], type[i]==1);
            }
            pl("");
        }
        for (int i=0; i<lineLen; i++)
            p("=");
        pl("");

        // Print totals
        for (int i=0; i<fieldCount; i++) {
            if (i>0)
                p(" | ");
            int len = maxLen[i];
            if (type[i] == 1)
                field(""+sum[i], maxLen[i], true);
            else
                field("", maxLen[i], false);
        }
        pl("");
        int size = results.size();
        pl(size + " row" + (size==1 ? "" : "s") + " found");
    } // run

    /** Pad str to field of length n (if n &gt; str.length()).
     ** Padding goes before if isInt, otherwise after.
     **/
    private void field(String str, int n, boolean isInt) {
        n -= str.length();
        if (n<=0) n = 0;
        if (isInt) while (n-- > 0) p(" ");
        p(str);
        if (!isInt) while (n-- > 0) p(" ");
    }
} // QueryTest
